/*
 * Addon.java
 *
 * An object identifying a specific addon, including all identifying information
 * regarding that addon, and providing a method with which to request its
 * update (via XML, TOC, or some other medium).
 *
 * Also implements a Runnable interface so that other addons can create a new
 * thread on this addon to get updated info about it, and then callback.
 *
 * Created on March 1, 2007, 4:22 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;

import java.io.*;
import java.net.*;
import java.util.concurrent.Callable;
import javax.swing.JTextField;
import org.xml.sax.helpers.XMLReaderAdapter;
import java.util.regex.*;
import java.util.Vector;

/**
 * An object identifying a specific addon, including all identifying information
 * regarding that addon, and providing a method with which to request its
 * update (via XML, TOC, or some other medium).
 *
 * Also implements a Runnable interface so that other addons can create a new
 * thread on this addon to get updated info about it, and then callback.
 *
 * @author m.delbuono
 */
public class Addon implements Runnable
{
        /** The UID of this addon, as identified at WoWInterface */
	protected int uid = -1;
        /** The name of this addon, as supplied by the author */
	protected String name = "";
        /** The name of the author of this addon */
	protected String author = "";
        /** A list of directories this addon uses at Interface/Addons */
	protected String[] directory;
        /** The MD5 hash of the archive available at WoWInterface */
	protected String md5 = "";
        /** The version string identifying this addon */
	protected String version = "";
        /** The URL at which this addon is accessible */
	protected String url;
	//protected String path;
        /** The filename for this addon */
        protected String file;
        /** In cases where a different MD5 hash has been detected, this
         * stores the new MD5 available from the patcher data */
        protected String newMD5 = null;
        /** A vector of Files which lists "dependents" (folders which
         * the updater does not care about directly, but are involved with
         * this addon).
         */
        protected Vector<File> dependents = new Vector<File>();
        /** The size, in bytes, of this addon */
        protected int size;
        
        private Callable c;
        private JTextField cfield;
        
        /** A file representing the TOC file for this addon */
        protected File toc;
        /*
        public File getTOC()
        {
            return toc;
        }
         */
        
        /** Gets the TOC file for this addon
         * @return a File object representing the TOC file for this addon 
         */
        public File GetTOC()
        {
            return toc;
        }

        /** Sets up the addon with only a valid UID. Call run() either manually
         * or via a Thread.start() to have the addon filled with correct
         * values from patcher data
         * @param   uid The UID to base this addon off of
         */
        public Addon(int uid)
        {
            this.uid = uid;
            c = null;
        }
        
        /** Converts a favorites entry into an addon 
         * @param   e   A <tt>FavoritesEntry</tt> to read data from
         */
        public Addon(FavoritesEntry e)
        {
            uid = e.uid;
            name = e.name;
            author = e.author;
            version = e.version;
        }

        /** Designed for AddonInstallerDialog, allows a callback method
         * for handling a web-based retrieval of information in a separate
         * thread from an event queue
         * @param   callback    A <tt>Callable</tt> to call after the process is complete
         * @param   cfield      A <tt>JTextField</tt> to update
         */
        public Addon(Callable callback, JTextField cfield)
        {
            this.cfield = cfield;
            c = callback;
        }
        
        /** Creates an addon from a File representing the directory
         * in which the addon is stored.
         *
         * @param   f   The directory to load the addon from
         * @throws FileNotFoundException if the addon was not found
         * @throws IOException if an I/O error occurred
         */
        public Addon(java.io.File f) throws FileNotFoundException, IOException
        {
		ConstructAddon(f);	
        }
        
        /** Creates a new addon from a String representing the path to the
         * directory in which the addon is stored.
         *
         * @param   path    the path to load the addon from
         * @throws FileNotFoundException if the addon was not found
         * @throws IOException if an I/O error occurred
         */
	public Addon(String path) throws FileNotFoundException, IOException
	{
		// Try to open the WID file related to this addon
		java.io.File p = new File(path);
		java.io.File f = new File(p, Main.DATA_FILE);
		
                ConstructAddon(f);
			
	}
        
        private void ConstructAddon(File f) 
            throws IOException, FileNotFoundException, IllegalArgumentException
        {
            if (!f.exists())
		throw new FileNotFoundException(Main.DATA_FILE+" not found");
            
            toc = new File(f.getParentFile(), f.getParentFile().getName()+".toc");
            
            if (!toc.exists())
                throw new FileNotFoundException(f.getParentFile().getName()+".toc not found");
            
            java.io.FileReader fr = new FileReader(f);
            java.io.BufferedReader in = new BufferedReader(fr);

            String buffer;
            // Read file data
            buffer = in.readLine();
            if (0 == buffer.compareToIgnoreCase("$$REGISTER"))
            {
                uid = -1;
                name = f.getParentFile().getName();
                author = "";
                md5 = "";
                version = "";
                url = "";
            }
            else if (0 == buffer.compareToIgnoreCase("$$DEPENDENT"))
            {
                throw new IllegalArgumentException("Addon Folder Dependent");
            }
            else
            {
                uid = Integer.parseInt(buffer);
                name = in.readLine();
                author = in.readLine();
                md5 = in.readLine();
                version = in.readLine();
                url = in.readLine();
            }
            
            
        }
        

	/** Attempts to update the addon by connecting to WoWI's download server
         */
	public boolean Update()
	{
            try
            {              
		HttpURLConnection socket = Connect();
                String data;

		// Grab the data into an XMLReader
		data = (String)socket.getContent();
                socket.disconnect();
                
                // First, let's find the md5 info and compare
                String currentMd5 = null;
               
                Pattern regex = Pattern.compile("<UIMD5>([a-fA-F0-9]+)</UIMD5>",
                        Pattern.CANON_EQ);
                Matcher matcher = regex.matcher(data);
                if (matcher.find())
                {
                    currentMd5 = matcher.group();
                    if (md5 == currentMd5)
                        return true;    // Nothing to update
                    else
                        return DoUpdate(data);
                }
                else
                {
                    throw new ProtocolException("Unable to match MD5 regex");
                }                    
            }
            catch (ProtocolException e) 
            {
                return false;
            }
            catch (MalformedURLException e)
            {
                FileReporterHandler.err("Internal error: Malformed URL getting patcher data");
                return false;
            }
            catch (IOException e)
            {
                return false;
            }
                
	}
        
        
        
        private boolean DoUpdate(String data)
        {
            return false;
        }

    /** Gets the name of the addon
     * @return a String representing the name of the addon
     */
    public String GetName() {
        return name;
    }
    /** Gets the UID representing this addon
     * @return the UID representing this addon
     */
    public int GetUID() {
        return uid;
    }
    /** Gets the author who wrote this addon
     * @return a String representing the Author who wrote this addon
     */
    public String GetAuthor() {
        return author;
    }
    
    /** Pulls whatever information it can from the TOC file into 
     * this object
     * @throws IOException if an I/O error occurred
     */
    public void UpdateFromTOC() throws IOException
    {
            
        // Pull up author's name
        Pattern authorPattern = Pattern.compile("##[ ]*Author: (.+)");
        FileReader fr = new FileReader(toc);
        BufferedReader reader = new BufferedReader(fr);
        String line;
        while ((line = reader.readLine())!= null)
        {
            Matcher m = authorPattern.matcher(line);
            if (m.find())
            {
                author = m.group(1);
                break;
            }
        }
        
        if (author.length() < 2)            // We didn't get a valid author
            // More and more authors are leaving ##AUTHOR out, so now we'll just
            // have to accept it and give it something like "unknown."
            author = "Unknown";
    }
    
    
    /** Pulls all information from XML patcher file<br />
     * <b>Note: </b>Does not update MD5! (But a future call to 
     * <tt>ConfirmInstall()</tt> will do so)
     *@param    d   The XML data to read
     *@throws IllegalArgumentException if the XML data is invalid
     */
    public void UpdateFromXML(String d)
        throws java.lang.IllegalArgumentException
    {
        this.name = ParseXML(d, "UIName");
        this.file = ParseXML(d, "UIFile");
        this.author = ParseXML(d, "UIAuthorName");
        this.version = ParseXML(d, "UIVersion");
        this.size = Integer.parseInt(ParseXML(d, "UISize"));
        this.url = ParseXML(d, "UIFileURL");
        this.newMD5 = ParseXML(d, "UIMD5");
    }
    
    /** Extracts the requested XML element into a string
     *@param    element the field to extract
     *@throws IllegalArgumentException if the requested element could not be found 
     *@return the requested element
     */
    private String ParseXML(String xml, String element)
        throws java.lang.IllegalArgumentException
    {
        Pattern p = Pattern.compile("<"+element+">(.+?)</"+element+">");
        Matcher m = p.matcher(xml);
        if (!m.find())
            throw new java.lang.IllegalArgumentException("XML Missing <"+element+"> element");
        return m.group(1);
    }
    
    /** Extracts the requested element from the TOC
     *@param    element the field to extract
     *@return the requested element
     *@throws IllegalArgumentException if the element requested could not be found
     */
    public String ParseTOC(String element)
        throws java.lang.IllegalArgumentException, 
            FileNotFoundException, IOException
    {
        // Check if we don't have a TOC, and handle that case
        if (toc == null)
        {
            return "Unknown (New Addon)";
                
        }
        
        Pattern p = Pattern.compile("##[ ]*"+element+":[ ]*(.+)",
                                 Pattern.CASE_INSENSITIVE | Pattern.CANON_EQ);
        FileReader fr = new FileReader(toc);
        BufferedReader in = new BufferedReader(fr);
        String line;
        
        while ((line = in.readLine())!= null)
        {
            Matcher m = p.matcher(line);
            if (m.find())
            {
                return m.group(1);
            }
        }
        
        throw new IllegalArgumentException("TOC element "+element+" not found");
    }
    
    /** Updates this addon's MD5 hash to the given MD5
     * @param   md5 a <tt>String</tt> representing this addon's new MD5
     */
    public void UpdateMD5(String md5)
    {
        this.md5 = md5;
    }
    
    /** Sets this addon's UID
     * @param   id  The addon's new UID
     */
    public void SetUID(int id)
    {
        uid = id;
    }
    
    /** Gets the URL to download this addon from
     * @return This addon's URL
     */
    public String GetURL()
    {
        return this.url;
    }
    
    /** Gets the path which this addon represents
     * @return the addon's path
     */
    public String GetFile()
    {
        return this.file;
    }
    
    /** Tests if the MD5 currently known matches that in the patcher
     * data xml.
     * @param   xml The XML data to read
     * @return <tt>true</tt> if the MD5 matches, <tt>false</tt> otherwise
     * @throws IllegalArgumentException if the XML data is bad
     */
    public boolean checkMD5FromXML(String xml)
        throws IllegalArgumentException
    {
        String newMd5 = ParseXML(xml, "UIMD5");
        if (newMd5.compareToIgnoreCase(this.md5) == 0)
            return true;
        else
            return false;
    }
    
    /** Gets the size of this addon archive
     * @return the size of this addon, in bytes
     */
    public int GetSize()
    {
        return size;
    }
    
    /** Gets the version string of this addon
     * @return the version string representing this addon
     */
    public String GetVersion()
    {
        return this.version;
    }
    
    /** Updates the MD5 in this addon, after a call to UpdateMD5
     * @see wowiupdater.Addon#UpdateMD5(String md5)
     */
    public void ConfirmInstall()
    {
        this.md5 = this.newMD5;
        this.newMD5 = null;
    }
    
    /** Marks this file as a dependent (meaning, while it may
     * be outside the main addon folder, it still is part of this addon
     * and should not be treated as its own)
     * This is used later when creating WIDs.
     *
     * @param   f   Path to mark as a dependent
     */
    public void addDependent(File f)
    {
        dependents.add(f);
    }
    
    /** Writes the WID file to the file specified
     * <code>
     * File format:
     *      UID
     *      Name
     *      Author
     *      MD5
     *      Version
     *      URL
     * </code>
     * @param   f   The file to write WID data to
     * @throws IOException if an I/O error occurred
     */
    public void writeWID(File f) throws IOException
    {
        FileWriter fw = new FileWriter(f);
        BufferedWriter out = new BufferedWriter(fw);
        out.write(String.valueOf(this.uid));
        out.newLine();
        out.write(this.name);
        out.newLine();
        out.write(this.author);
        out.newLine();
        out.write(this.md5);
        out.newLine();
        out.write(this.version);
        out.newLine();
        out.write(this.url);
        out.flush();
        out.close();
        fw.close();
        
        
        // Now let's write out dependency lines for anything that is 
        // NOT this folder
        
        // TODO: Something seems strange here. Are we really traversing
        // all dependents?
        for (int x = 0; x < dependents.size(); x++)
        {
            if (!dependents.get(x).isDirectory())
                continue;                   // Don't do this on root files
            
            File depFile = new File(dependents.get(x), f.getName());
            if (f.equals(depFile))        // Don't break up this one!                
                continue;           
            
            FileWriter writer = new FileWriter(depFile);
            BufferedWriter os = new BufferedWriter(writer);
            os.write("$$DEPENDENT");
            os.flush();
            os.close();
        }
    }
    
    private HttpURLConnection Connect() throws MalformedURLException, IOException
    {
        String surl = String.format(Main.WOWI_URL + Main.UPDATE_FORMAT, uid);
        URL url = new URL(surl);
        
        FileReporterHandler.verbose("I++ "+Thread.currentThread().getName()
            +" attempts to connect to "+surl);
        
        HttpURLConnection socket = (HttpURLConnection)url.openConnection();
        socket.setFollowRedirects(true);
        socket.setRequestMethod("GET");
        socket.connect();
        return socket;
                
    }

    /** Pulls in info from the UID */
    public void run() {
        HttpURLConnection socket = null;
        try
        {
            try
            {
                FileReporterHandler.verbose("W?? "+Thread.currentThread().getName()+
                        " requests lock on 'this' (Addon)");
                synchronized (this)
                {
                    FileReporterHandler.verbose("W++ "+Thread.currentThread().getName()+
                            " gains lock on 'this' (Addon)");
                    this.wait(500); // Slight delay to allow for update
                }
                FileReporterHandler.verbose("W-- "+Thread.currentThread().getName()+
                        " releases lock on 'this' (Addon)");

                if (cfield != null)
                    uid = Integer.parseInt(cfield.getText());

                socket = Connect();
                String xml;
                xml = inStreamToString((InputStream)socket.getContent());
                UpdateFromXML(xml);
                if (!Thread.currentThread().isInterrupted() && c != null)
                    c.call();
            }
            catch (InterruptedException e)
            {
                FileReporterHandler.verbose("W## "+Thread.currentThread().getName()+
                        " interrupted in wait loop - releasing lock");
                // Ignore - interruption is ok
            }
            catch (NumberFormatException e)
            {
                // Also ok, just can't set any info
                if (c != null)
                    c.call();
            }
            catch (IllegalArgumentException e)
            {
                // Bad file, just send blank
                if (c != null)
                    c.call();
            }
        }
        catch (Exception e)
        {
            FileReporterHandler.err("Uncaught exception in Addon.run()");
            FileReporterHandler.dump(e);
        }
        finally
        {
            if (socket != null)
                socket.disconnect();
        }
    }

    // Creates a String from the given input stream and then closes it
    private String inStreamToString(InputStream in) throws IOException {
        String out = "";
        int b;
        while ((b = in.read())!=-1)
            out += (char)b;
        
        return out;
    }
    
    /** Converts this addon into a string representation, which helps
     * with listings and other such generic systems.
     * @return the String representation of this object
     */
    public String toString()
    {
        return this.name + " by " + this.author + " (" +this.uid+")";
    }

    String GetServerHash() {
        return md5;
    }

}
