/*
 * AddonExtractor.java
 *
 * Extracts an addon from a archive format (ZIP, RAR, etc.). This
 * encapsulates the various extraction methods that would be required, so
 * the updater is not dependent upon file formats. New formats can
 * be quickly added.
 *
 * Created on March 30, 2007, 3:06 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 *
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;

import java.io.*;
import java.util.Enumeration;
import java.util.regex.Pattern;
import java.util.zip.*;

/**
 * Extracts an addon from a archive format (ZIP, RAR, etc.). This
 * encapsulates the various extraction methods that would be required, so
 * the updater is not dependent upon file formats. New formats can
 * be quickly added.
 *
 * @author m.delbuono
 */
public class AddonExtractor {
    private File file = null;
    private int filetype;
    
    //public static final int TYPE_UNKNOWN = 0;
    /** Identifies the archive as a ZIP type */
    public static final int TYPE_ZIP = 1;
    /** Identifies the archive as a RAR type */
    public static final int TYPE_RAR = 2;
   
    // Used for writing
    private FileOutputStream fos = null;
    private OutputStream out = null;
    
    // Used for reading (wtb C unions :( ) 
    private Enumeration listing = null;
    private InputStream in = null;
    private ZipFile infile = null;
    private ZipEntry nextEntry = null;
    
    /** Creates a new instance of AddonExtractor 
     * @param   f   the file to extract or write
     */
    public AddonExtractor(File f) {
        // Get file type
        file = f;
        String ext = getExtension();
        FileReporterHandler.verbose("Z!! "+Thread.currentThread().getName()+
            " Identified archive type: "+ext);
        if (ext.compareToIgnoreCase(".zip")==0)
        {
            filetype = TYPE_ZIP;
        }    
        else if (ext.compareToIgnoreCase(".rar")==0)
        {            
            filetype = TYPE_RAR;
        }
        else
        {
            throw new UnsupportedOperationException("Unknown archive format");        
        }
    }
    
    /** Creates a new AddonExtractor, forcing a specific archive format
     * @param   f   The file to extract or write
     * @param   type    The type of archive to read or write
     */
    public AddonExtractor(File f, int type)
    {
        if (!(type == TYPE_ZIP || type == TYPE_RAR))
            throw new UnsupportedOperationException("Unknown archive format");
            
        file = f;
        filetype = type;        
    }
    
    /** Opens the archive ready for writing
     * @throws IOException if an error occurred while opening
     */
    public void OpenForWriting() throws IOException
    {
        Close();
        
        if (filetype == TYPE_ZIP)
        {
            FileOutputStream fos = new FileOutputStream(file);
            out = new ZipOutputStream(fos);
        }
        else if (filetype == TYPE_RAR)
            throw new UnsupportedOperationException("Writing of RAR files is NYI");
    }
    
    /** Opens the archive ready for reading
     * @throws IOException if an error occurred during opening
     */
    public void OpenForReading() throws IOException
    {
        Close();
        
        if (filetype == TYPE_ZIP)
        {
            infile = new ZipFile(file);
            listing = infile.entries();
        }
        else 
            throw new UnsupportedOperationException("Reading of archives "+
                    "of type "+filetype+" is NYI");
    }
    
    /** Closes all streams
     * @throws IOException if an error occurred while closing
     */
    public void Close() throws IOException
    {
        if (out != null)
            out.close();
        if (in != null)
            in.close();
        if (fos != null)
            fos.close();
        if (infile != null)
            infile.close();
        
        
        out = null;
        in = null;
        fos = null;
        infile = null;
    }
    
    /** Adds a file to the archive
     * @param   f   The file to add
     * @param   path    The path to add the file to
     * @throws IOException if an error occurred while adding
     */
    public void add(File f, String path) throws IOException
    {       
        if (out == null)
            throw new IOException("Archive is not ready for writing");
        
        if (filetype == TYPE_ZIP)
        {
            // So we don't have to typecast everything...
            ZipOutputStream zip = (ZipOutputStream)out;
            // Make a new zip entry
            ZipEntry ze = new ZipEntry(path);
            // Add it
            zip.putNextEntry(ze);
            if (f.isFile())
            {
                // Make a reader
                FileInputStream in = new FileInputStream(f);
                int b;                      // stores a byte
                while ((b = in.read())!=-1) // While there's more...
                    zip.write(b);              
            }
            // Finalize
            zip.closeEntry();
        }
        else
        {
            throw new UnsupportedOperationException("Adding to archive type "+
                    filetype+" is not yet implemented");
        }
            
    }
    
    
    /** Prepares the next element.
     * @return <tt>true</tt> if there are more elements to read, <tt>false</tt> if not (or if not open for reading)
     */
    public boolean next()
    {
        if (infile == null)
            return false;
        
        if (filetype == TYPE_ZIP)
        {
            if (!listing.hasMoreElements())
                return false;
            
            nextEntry = (ZipEntry)listing.nextElement();
            return true;                
        }
        else
        {
            throw new UnsupportedOperationException("Checking element status "+
                    "of archive type "+filetype+" is NYI");
        }
    }

    /** Determines if the current entry is a directory
     * @return <tt>true</tt> if the entry is a directory, <tt>false</tt> if not
     */
    public boolean entryIsDirectory() throws IOException {
        if (filetype == TYPE_ZIP)
        {
            if (nextEntry == null)
                throw new IOException("No entry");
            
            return nextEntry.isDirectory();
        }
        else
            throw new UnsupportedOperationException("Getting entries for "+
                    "archive type "+filetype+" is NYI");
    }

    /** Gets the current entry's name
     * @return the name of the current entry
     * @throws IOException if there is no current entry
     */
    public String entryName() throws IOException 
    {
        if (filetype == TYPE_ZIP)
        {
            if (nextEntry == null)
                throw new IOException("No entry");
            
            return nextEntry.getName();
        }
        else 
            throw new UnsupportedOperationException("Getting entries for "+
                    "archive type "+filetype+" is NYI");
    }
    
    /** Determines true if the current entry's name matches a supplied pattern
     * @param   pattern A regex pattern to test
     * @param   caseInsensitive <tt>true</tt> if performing a case insensitive search, <tt>false</tt> if not
     * @return <tt>true</tt> if the entry matches the pattern, <tt>false</tt> if not
     */
    public boolean entryMatches(String pattern, boolean caseInsensitive) 
        throws IOException 
    {   
        Pattern p = Pattern.compile(pattern, Pattern.CANON_EQ | 
                (caseInsensitive?Pattern.CASE_INSENSITIVE:0)
        );
        //p.matcher(pattern);
        return p.matcher(entryName()).find();
    }
    
    /** Returns the total number of entries in the file
     * @return the total number of entries in the file
     */
    public int size() throws IOException
    {
        if (filetype == TYPE_ZIP)
        {
            if (infile == null)
                throw new IOException("Input stream is closed");
            else
                return infile.size();
        }
        else
            throw new UnsupportedOperationException("Getting size of a type "+
                    filetype+" archive is NYI");
    }
    
    /** Writes the current entry to the specified file
     * @param   f   The file to which the entry should be written
     * @throws IOException if an I/O error occurred
     */
    public void write(File f) throws IOException
    {
        if (filetype == TYPE_ZIP)
        {
            if (infile == null || nextEntry == null)
                throw new IOException("No entry");
            
            // Make an output stream on the file
            BufferedOutputStream os = new BufferedOutputStream(
                    new FileOutputStream(f)
            );
            in = infile.getInputStream(nextEntry);
            
            // Now write
            int b;
            while ((b = in.read())!=-1)
                os.write(b);
            os.flush();
            os.close();
            in.close();
        }
        else
            throw new UnsupportedOperationException("Reading entries from "+
                    "type "+filetype+" archives is NYI");
    }

    /** Gets the extension of the archive file
     * @return the extension of the file, such as ".zip"
     */
    public String getExtension()
    {
        return file.getName().substring(file.getName().length()-4);
    }
    
}
