/*
 * AddonReverter.java
 *
 * Reverts an addon given a backup archive to use, after creating a backup
 * before the reversion.
 *
 * Created on April 6, 2007, 11:01 AM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 *
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Vector;
import javax.swing.*;
import java.awt.*;
/**
 * Reverts an addon given a backup archive to use, after creating a backup
 * before the reversion.
 *
 * @author m.delbuono
 */
public class AddonReverter implements Runnable {
    File zip;
    BackupEntry entry;
    boolean processSV;
    JTextField status;
    JProgressBar progress;
    JButton done;
    Addon addon;
    /** Creates a new instance of AddonReverter */
    public AddonReverter(BackupEntry backup, boolean processSavedVars,
            JTextField status, JProgressBar progress, JButton doneButton) 
    {
        processSV = processSavedVars;
        zip = backup.location;
        entry = backup;
        this.status = status;
        this.progress = progress;
        this.done = doneButton;
        this.addon = addon;
    }
    
    /** Sets the current status output:
     * <ul>
     * <li>If <tt>text</tt> is <tt>null</tt>, leaves it unchanged</li>
     * <li>If <tt>start</tt> and/or <tt>end</tt> is <0, leaves <tt>start</tt> 
     * and/or <tt>end</tt> unchanged, respectively</li>
     * <li>If <tt>end</tt> is 0, sets indeterminate mode</li>
     * <li>If <tt>val</tt> is -1, leaves it unchanged. If <tt>val</tt> is -2, 
     * steps <tt>val</tt> by 1</li>
     * </ul>
     * @param text the text to display on the status bar
     * @param start the starting point of the progress bar
     * @param end the ending point of the progress bar
     * @param val the value of the progress bar
     */
    private void SetStatus(String text, int start, int end, int val)
    {
        if (text != null)
            status.setText(text);
        if (start > -1)
            progress.setMinimum(start);
        
        if (end == 0)
            progress.setIndeterminate(true);
        else if (end > -1)
        {
            progress.setIndeterminate(false);
            progress.setMaximum(end);
        }
        
        if (val == -2)
            progress.setValue(progress.getValue()+1);
        else if (val > -1)
            progress.setValue(val);
        
    }

    /** Runs the reversion */
    public void run() {
        try
        {
            done.setEnabled(false);
            
            final File addonsDir = new File(Main.wowPath, Main.REL_ADDONS);
            final File svDir = new File(Main.wowPath, Main.REL_SV);
            // First, we have to make a backup of where we currently stand
            SetStatus("Preparing for backup before reversion...", 0, 0, 0);

            // First make a list of files that we'll need to backup
            Vector<File> files = new Vector<File>();
            AddonExtractor zipExtractor = new AddonExtractor(zip);
            zipExtractor.OpenForReading();
            while (zipExtractor.next())
            {
                String path = zipExtractor.entryName();
                if (!zipExtractor.entryIsDirectory())
                {
                    // To be backed up, this must be a SV or addon
                    File f = new File(Main.wowPath.getParentFile(), path);
                    if (AddonUpdateProcess.isChild(f, svDir) ||
                        AddonUpdateProcess.isChild(f, addonsDir))
                        files.add(f);
                }
            }
            zipExtractor.Close();
            
            // Great, we got a list of directories. Now use them to backup
            SetStatus("Backing up before reversion...", 0, files.size(), 0);
            String backupFileName = "backup-rev-"+Main.filetime()+".zip";
            File backupFile = new File(Main.wowPath, Main.REL_BACKUPS+backupFileName);
            backupFile.createNewFile();
            AddonExtractor backup = new AddonExtractor(backupFile);
            backup.OpenForWriting();
            for (int i = 0; i < files.size(); i++)
            {
                File thisFile = files.get(i);
                // We only backup what's in interface/addons or SVs, remember
                if (!(AddonUpdateProcess.isChild(thisFile, addonsDir) ||
                      AddonUpdateProcess.isChild(thisFile, svDir))
                   )
                    continue;   // skip if outside of scope
                
                // Do the backup. List files, etc.
                backup.add(thisFile, 
                        AddonUpdateProcess.getRelativePath(
                             thisFile, Main.wowPath, '/'
                        )
                );
                
                
                SetStatus(null, -1, -1, -2);
            }
            backup.Close();
            
            Main.backupManager.OpenForAppending();
            SimpleDateFormat fmt = new SimpleDateFormat(
                //  2 Jul 2007 1:06 PM
                   "d MMM yyyy h:mm a"
                );
            String dt = fmt.format(new Date(entry.timestamp));
            Main.backupManager.AddBackup(backupFile, addon, "Restored backup "+
                    "from "+dt);
            Main.backupManager.Close();
            
            
            // Now clean out those directories in preparation for install
            SetStatus("Cleaning install location...", 0, files.size(), 0);
            for (int i = 0; i < files.size(); i++)
            {
                // In all cases, we'll delete the file
                if (files.get(i).exists())
                    files.get(i).delete();
                
                // For folders, only clean out addon folders, for safety
                if (AddonUpdateProcess.isChild(
                        files.get(i).getParentFile(),
                        addonsDir)
                )
                {
                    AddonUpdateProcess.CleanDirectory(
                            files.get(i).getParentFile(), true);
                }
                SetStatus(null, -1, -1, -2);
            }
            
            
            // So we've done the backup. Now replace!
            zipExtractor.OpenForReading();
            SetStatus("Restoring backup files...", 0, files.size(), 0);
            while (zipExtractor.next())
            {
                // Build the file name
                File installLocation =
                        new File(
                            Main.wowPath.getParentFile(),
                            zipExtractor.entryName()
                        );
                // Make the directory for it if needed
                installLocation.getParentFile().mkdirs();
                installLocation.createNewFile();
                // Now install the file
                try
                {
                    zipExtractor.write(installLocation);    
                }
                catch (IOException e)
                {
                    // There are certain cases where we can't restore
                    // (access denied, etc.) just ignore them
                }
                SetStatus(null, -1, -1, -2);
            }            
            zipExtractor.Close();
            
            // All done!
            SetStatus("Restore Complete", 0, 1, 1);
            done.setEnabled(true);
            
        }
        catch (Throwable e)
        {
            SetStatus("Error - Internal Error Occurred", 0, 1, 0);
            done.setEnabled(true);
            Thread.currentThread().getDefaultUncaughtExceptionHandler().uncaughtException(
                    Thread.currentThread(), e);
        }
        
        
    }
    
}
