/*
 * AddonUpdateProcess.java
 *
 * A class identifying an addon in question and abstracting the 
 * entire update process. This includes, but is not limited to,
 * suspension when user input is required, holding thread locks,
 * and downloading appropriate information.
 *
 * Created on March 1, 2007, 5:01 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.concurrent.CancellationException;
import java.util.logging.FileHandler;
import javax.swing.*;
import java.awt.*;
import java.io.*;
import java.util.regex.*;
import java.util.Vector;
import java.net.*;
//import java.util.zip.*;

/**
 * A class identifying an addon in question and abstracting the 
 * entire update process. This includes, but is not limited to,
 * suspension when user input is required, holding thread locks,
 * and downloading appropriate information.
 *
 * @author m.delbuono
 */
// Abstracts the Addon's update system
public class AddonUpdateProcess implements Runnable {
    /** A constant indicating the location of the SavedVariables folder */
    public static String SV_BASE_LOCATION = "SavedVariables";
    
    private Addon addon;
    private String status;
    private int progressStart, progressEnd, progressVal;
    private JProgressBar bar;
    private JLabel label;
    private GridBagConstraints frameConstraint;
    private Thread thread;      // The thread this process runs on
    private boolean running = false;
    private JComponent[] row;
    
    private Vector<File> savedVars = null;
    private Vector<File> savedVarsPerChar = null;
    
    private Vector<String> done = new Vector<String>();
    
    public static final Color
            SUSPENDED = new Color(1,1,0.8f),
            COMPLETE = new Color(0.8f,0.8f,1),
            ERROR = new Color(1,0.8f,0.8f),
            STANDARD = new Color(1,1,1);
    
    
    //public boolean leaseRequested = false;
    
    /** Creates a new instance of AddonUpdateProcess */
    public AddonUpdateProcess(Addon a) {
        this.SetAddon(a);        
    }
    
    /** Sets the background that this process should update 
     * @param row The component whose background will change
     */
    public void SetBackground(JComponent[] row)
    {
        this.row = row;
        SetBackground(STANDARD);
    }
    
    /** Sets the addon which this process will handle
     * @param a the addon to update
     */
    public void SetAddon(Addon a)
    {
        addon = a;
        status = "Waiting for available thread...";
        progressStart = 0;
        progressEnd = 1;
        progressVal = 0;
    }
    
    /** Sets the background of the current controlled component
     * @param c the <tt>Color</tt> to set
     */
    public void SetBackground(Color c)
    {
        row[0].setBackground(c);
        row[1].setBackground(c);
        row[2].setBackground(c);
        
        for (int i = 0; i < 3; i++)
            row[i].repaint();
    }
    /*
    public int GetProgressStart()
    {
        return progressStart;
    }
    public int GetProgressEnd()
    {
        return progressEnd;
    }
    public int GetProgress()
    {
        return progressVal;
    }
    public String GetStatus()
    {
        return status;
    }*/
    /** Gets the name of the addon this process handles
     * @return the name of this addon
     */
    public String GetName()
    {
        return addon.GetName();
    }
    
    /** Sets the progress bar related to this process
     * @param b the progress bar to control
     */    
    public void SetProgressBar(JProgressBar b)
    {
        bar = b;
        b.setMinimum(progressStart);
        b.setMaximum(progressEnd);
        b.setValue(progressVal);
    }
    
    /** Sets the controlling constraint for this addon
     * @param c a <tt>GridBagConstraints</tt> to handle component constraints
     */
    public void SetProgressConstraint(GridBagConstraints c)
    {
        frameConstraint = c;
    }
   
    /** Sets the label which will control status information
     * @param l the label to control
     */
    public void SetStatusText(JLabel l)
    {
        label = l;
        l.setText(status);
    }
    
    //@todo: Do we really need this?
    /*public void SetThread(Thread t)
    {
        thread = t;
    }
    public void GetThread()
    {
        return thread;
    }*/
    
    /** Determines if this process has completed
     * @return <tt>true</tt> if this process is currently running, or 
     * <tt>false</tt> if it has completed
     */
    public boolean isRunning()
    {
        return running;
    }
    
    /** Updates the status panels. If <tt>val</tt> is -1, it indicates the progress
     * bar is to enter an indeterminate state. If <tt>min</tt> or <tt>max</tt> are -1, then
     * it indicates the minimum or maximum values of the progress bar
     * are not to change, respectively. -2 for <tt>value</tt> indicates not to change
     * the value. -3 for <tt>value</tt> indicates to increment by 1.
     * @param s A string indicating the current status
     * @param val the value to set
     * @param min the minimum value
     * @param max the maximum value
     * @param c the current status color
     */
    public void UpdateStatus(String s, int val, int min, int max, Color c)
    {
        //JProgressBar pbar = (JProgressBar)bar;
        if (s != null) label.setText(s);
        bar.setIndeterminate((val==-1)?true:false);
        
        if (val >= 0)
            bar.setValue(val);
        
        if (val == -3)
            bar.setValue(bar.getValue()+1);
        
        if (min != -1)
            bar.setMinimum(min);
        
        if (max != -1)
            bar.setMaximum(max);
        /*
        SetBackground(c);
         */
        
    }
    
    /** The main routine, fired when entering its own thread, which allows
     * the system to start updating this addon and updates the MainFrame
     * accordingly.
     */
    public void run()
    {
        running = true;
        UpdateStatus("Reading configuration...", -1, -1, -1, STANDARD);
        /** DEBUG CODE
        if (addon.GetName().compareToIgnoreCase("!OmniCC")==0)
        {
            Thread t = new Thread(new ProgressBarTester(bar));
            t.setDaemon(true);
            t.start();
        }*/
        /*
        while(true) // DEBUG: Infinite loop
        {
           
            Main.f.repaint();
        }
         
        */
        try
        {
            if (addon.GetUID() == -1)
            {
                /*
                UpdateStatus("Waiting for favorites list lease...", -1, -1, -1);
                this.leaseRequested = true;
                while (Main.favoritesLock != this)
                {
                    Main.f.repaint();       // Give some wait time
                }*/


                UpdateStatus("Reading TOC...", -1, -1, -1, STANDARD);
                try
                {
                    addon.UpdateFromTOC();
                }
                catch (IOException e)
                {
                    UpdateStatus("Error - Invalid Addon - TOC unreadable", 0, 0, 1, ERROR);
                    FileReporterHandler.err("Unable to read TOC for "+addon.GetName());
                    FileReporterHandler.dump(e);
                    return;                 // Cancel this thread
                }

                // Now that we have the author, we can try to match to our favorites
                Vector<FavoritesEntry> matches = 
                    Main.favorites.match(this.addon.GetName(), this.addon.GetAuthor());

                if (matches.size() == 1 && Main.favorites.size() > 1)
                {
                    // We assume if there's only one match, and there is more than 
                    // one in Main.favorites, that this must be correct
                    addon.uid = matches.get(0).uid;
                }
                else if (matches.size() == 0)
                {
                    try
                    {
                        GetUID();
                    }
                    catch (CancellationException e)
                    {
                        return;
                    }
                }
                else
                {
                    CheckUID(matches);
                    //throw new UnsupportedOperationException("Not yet implemented");
                }
            }

            CheckInterruption();
            // So we (hopefully) got the UID. Now let's see what we can do
            // about it!
        
        
            // We have a XX step process:
            /*  Step 1: Retrieve patcher data
             *  Step 2: Compare version info
             *  Step 3: Download zip if necessary
             *  Step 4: Request user confirmation
             *  Step 5: Backup old UI
             *  Step 6: Unzip new UI
             *  Step 7: Write up a new WID
             */
            boolean repeat = false;
            do
            {
                // Step 1: Retrieve patcher data
                UpdateStatus("Retrieving patcher data", -1, -1, -1, STANDARD);
                URL downloadUrl = GetDownloadURL();
                CheckInterruption();

                // Step 2: Compare versioning info
                if (downloadUrl == null)    // Matched
                {
                    UpdateStatus("Addon is up-to-date", 1, 0, 1, COMPLETE);
                    return;                 // Kills the thread, releases to newer
                }
                CheckInterruption();

                // Step 3: Download data! (The fun part!)
                File zip;
                if (addon.GetTOC() != null)
                {
                    zip = new File(
                            addon.GetTOC().getParentFile(),
                            addon.GetFile()
                    );
                }
                else
                {
                    zip = new File(
                            new File(Main.wowPath, Main.REL_ADDONS), 
                            addon.GetFile()
                    );
                }
                if (zip.exists()) zip.delete();
                zip.createNewFile();
                DoDownload(zip);
                CheckInterruption();

                // Step 4: Request user confirmation
                // We do this by showing a confirmation dialog comparing
                // TOC data to XML data
                if (!ShowConfirmation())
                {
                    repeat = true;
                    // Get a new UID from user
                    try
                    {
                        GetUID();
                        continue;
                    }
                    catch (CancellationException e)
                    {
                        return;
                    }
                }

                // Step 5: Create a backup (zip old addon, store in backup folder)
                // REMOVED: This is now done as part of step 6

                /*
                File backupFile = new File(addon.GetTOC().getParent(), 
                                            "backup_"+Main.filetime()+".zip");
                if (backupFile.exists())
                {
                    UpdateStatus("Error - Cannot create backup (code 1)", 0, 0, 1);
                    return;
                }

                try
                {
                    backupFile.createNewFile();
                    CreateBackup(backupFile);
                }
                catch (FileNotFoundException e)
                {
                    UpdateStatus("Error - Cannot create backup (code 2)", 0, 0, 1);
                    return;
                }
                catch (IOException e)
                {
                    UpdateStatus("Error - Cannot create backup (code 3)", 0, 0, 1);
                    return;
                }
                */
                // Finally... step 6, remove the old UI and install new
                UpdateStatus("Installing update...", -1, -1, -1, STANDARD);
                // Perform one last check for dependency
                CheckInterruption();
                zip = UIReplace(zip);


                // Now Step 7, rewrite the WID
                UpdateStatus("Finalizing installation...", -1, -1, -1, STANDARD);
                addon.ConfirmInstall();
                // If this is a new addon, zip.getParent() will be
                // the Interface/Addons directory. We need to push this
                // to the correct folder.
                File widFile = new File(zip.getParent(), Main.DATA_FILE);
                if (widFile.exists()) widFile.delete();
                widFile.createNewFile();
                addon.writeWID(widFile);
                UpdateStatus("Update complete. Current version: "+
                        addon.GetVersion(),1,0,1,COMPLETE);
            } while (repeat);
        }
        catch (MalformedURLException e)
        {
            UpdateStatus("Error - Malformed URL", 0, 0, 1,ERROR);
            FileReporterHandler.dump(e);
            return;
        }
        catch (ProtocolException e)
        {
            UpdateStatus("Error - Bad content retrieved", 0, 0, 1,ERROR);
            FileReporterHandler.err("Terminating thread '"+Thread.currentThread().getName()+"': Bad Content Retrieved");
            FileReporterHandler.dump(e);
            return;
        }
        catch (IOException e)
        {
            UpdateStatus("Error - Unspecified I/O Error", 0, 0, 1,ERROR);
            FileReporterHandler.err("Terminating thread '"+Thread.currentThread().getName()+"': Unspecified I/O Error");
            FileReporterHandler.dump(e);
            return;
        }
        catch (InterruptedException e)
        {
            FileReporterHandler.out(">> Caught IRQ-1, terminating thread "+Thread.currentThread().getName());
            running = false;
            return;
        }
        catch (UnsupportedOperationException e)
        {
            UpdateStatus("Error - Unsupported Addon", 0, 0, 1, ERROR);
            FileReporterHandler.err("Terminating thread '"+
                    Thread.currentThread().getName()+"': "+
                    "Unsupported operation");
            FileReporterHandler.dump(e);
            return;
        }
        catch (Throwable t)
        {
            // Only catching so we can update status and then throw out
            // (Just so user isn't confused when nothing happens)
            UpdateStatus("Error - Internal Error", 0, 0, 1, ERROR);
            Thread.currentThread().getDefaultUncaughtExceptionHandler().uncaughtException(Thread.currentThread(), t);
            return;
        }
        
        
        //UpdateStatus("Update successful!", 1, 0, 1);
        running = false;
        
    }
    
    /** Phase 1 of update process
     * @return URL the download location of the updated version, or null if current MD5s match
     */
    private URL GetDownloadURL() 
        throws MalformedURLException, ProtocolException, IOException
    {
        URL patcherUrl = new URL(Main.WOWI_URL + 
                String.format(Main.UPDATE_FORMAT, addon.GetUID()));

        URLConnection connection = patcherUrl.openConnection();
        
        Object content = connection.getContent();
        if (!(content instanceof InputStream))
            throw new ProtocolException("Bad content retrieved");
        
        int bytes = connection.getContentLength();
        if (bytes > 0)
            UpdateStatus("Retrieving patcher data", 0, 0, bytes, STANDARD);
        else if (bytes == 0)
            throw new ProtocolException("No content retrieved");
        else
            UpdateStatus("Retrieving patcher data", -1, -1, -1, STANDARD);
        
        InputStream stream = (InputStream)content;
        String xml = ""; int bt, bytes_read = 0;
        while ((bt = stream.read())!=-1)
        {
            xml += (char)bt;
            UpdateStatus("Retrieving patcher data", ++bytes_read, -1, -1, STANDARD);
        };
        
        
        // While we're here, let's update info
        try
        {
            addon.UpdateFromXML(xml);
        }
        catch (IllegalArgumentException e)
        {
            throw new ProtocolException();
        }
        
        // Now that we've got that, head on out!        
        // Returning null if they matched
        if (addon.checkMD5FromXML(xml))
            return null;
        else
            return new URL(addon.GetURL());
    }
    
    /** Downloads the file in question from the server
     * @param zip the file to download
     * @throws IOException if an I/O error occurred
     */
    private void DoDownload(File zip) throws IOException
    {
        UpdateStatus("Preparing for download...", -1, 0, addon.GetSize(), STANDARD);
        OutputStream out = new BufferedOutputStream(new FileOutputStream(zip));
        int b, i=0;
        
        URL url = new URL(addon.GetURL());
        URLConnection connection = url.openConnection();
        InputStream in = (InputStream)connection.getContent();
        
        // We'll build the MD5 while downloading to check
        MessageDigest digest;
        try
        {
            digest = MessageDigest.getInstance("MD5");
        }
        catch (NoSuchAlgorithmException e)
        {
            throw new IOException("Halted Download - Cannot validate MD5 hash");
        }
        
        
        while ((b=in.read()) != -1)
        {
            i++;
            UpdateStatus("Downloading... ( "+
                    java.lang.Math.floor(i/1024)+"Kb / "+
                    java.lang.Math.floor(addon.GetSize()/1024) + "Kb )",
                    i, -1, -1, STANDARD
            );
            out.write((byte)b);
            digest.update((byte)b);
        }
        in.close();
        out.close();
        
        
        // Now we want to search for dependents and alert early
        AlertDependents(zip);
        
    }
    
    
    /** Confirms that the file we downloaded is really what we want
     * @return <tt>true</tt> if the user confirmed the download, or <tt>false</tt>
     * if denied the download
     * @throws InterruptedException if the thread is interrupted while
     * in a wait loop
     */
    private boolean ShowConfirmation() throws InterruptedException
    {
        UpdateStatus("New version "+addon.GetVersion()+" available! "+
                "Awaiting confirmation...", -2, -1, -1, SUSPENDED);
        // First we set up a button and wait for them to click it
        DownloadConfirmationSuspender s =
                new DownloadConfirmationSuspender(addon);
        s.Suspend(frameConstraint, Main.f.getComponentPane(), bar);
        return s.isOkToInstall();
        
    }
    
    
    
    /** Creates a backup of the specified directory */
    private void CreateBackup(File f, File directory)
        throws FileNotFoundException, IOException
    {
        //UpdateStatus("Creating backup of old UI...", -1, -1, -1);
/*****************************'
 *CONVERTED TO ENCAPSULATOR
        FileOutputStream fos = new FileOutputStream(f);
        ZipOutputStream zip = new ZipOutputStream(fos);

 **********************************************/
        AddonExtractor zip = new AddonExtractor(f);
        // Figure out all the directories (note this will have to become recursive)
        //File directory = addon.GetTOC().getParentFile();
        File[] entries = directory.listFiles(new NonZipFilter());
        //ZipEntries(zip, entries, "");
        ZipEntries(zip, entries, "");
        try
        {
/*****************************
            zip.close(); 
            fos.close();
 ***************************/
            zip.Close();
        }
        catch (IOException e)
        {
            // Occasionally we will make an empty zip... just ignore em
            FileReporterHandler.err("Ignoring the following error: ");
            FileReporterHandler.dump(e);
        }
    }
    
    /** Recursive function to handle zipping all entries together, recursively
     * through subdirectories
     */
    private void ZipEntries(AddonExtractor zip, File entries[], String prefix
            /*Vector<String> done*/)
        throws IOException
    {
        /*
        if (done == null)
            done = new Vector<String>();
        */
        for (int i = 0; i < entries.length; i++)
        {
            // Build the path string (WHY SUN WHY CAN'T WE JUST USE A FILE!)
            String path = prefix;
            path += entries[i].getName();
            if (entries[i].isDirectory())
                path += "/";            // Directories must always end with /
            
            // Optimization - skip any entries we've already done
            if (done.contains(path))
            {
                FileReporterHandler.verbose("Z-- "+Thread.currentThread()+" skips entry "+path+" (already done)");
                continue;
            }
            else
                done.add(path);
            
            // Make the entry
/************************************************
 *CONVERTED TO ENCAPSULATOR
            ZipEntry ze = new ZipEntry(path);
            
            // Now add it in
            FileReporterHandler.verbose("Z<< "+Thread.currentThread()+" backs up "+ze.getName());
            if (prefix.equals(""))
                UpdateStatus("Backing up interface...",i,0,entries.length,STANDARD);
           
            zip.putNextEntry(ze);
            if (entries[i].isFile())
            {
                FileInputStream in = new FileInputStream(entries[i]);
                int b;
                while ((b = in.read())!=-1)
                {
                    zip.write(b);
                }
            }
            zip.closeEntry();
 *****************************************************/
            FileReporterHandler.verbose("Z<< "+Thread.currentThread()+
                    " backs up "+path);
            if (prefix.equals(""))
                UpdateStatus("Backing up interface...",i,0,entries.length,STANDARD);
            zip.add(entries[i], path);
            /*
            // Now we need to enumerate and recurse if this was a directory
            if (entries[i].isDirectory())
            {
                File[] newList = entries[i].listFiles(new NonZipFilter());
                ZipEntries(zip, newList, path, done);
            }*/
        }
    }
    
    /** Extracts the ZIP specified, attempting to detect the appropriate
     * directories. Any secondary addons found in this directory
     * will be marked as dependents upon this one (dependents
     * do not appear in the update routine) <br /><br />
     *
     * Also backs up and deletes any old components (will delete any non-zip
     * in base folder, and anything in subfolders)<br /><br />
     *
     * This is... most definitely... the hard part<br /><br />
     * 
     * <b>Algorithm (ordered):</b>
     * <ul>
     * <li>If "addons" folder is found, match it to WoW/Interface/Addons</li>
     * <li>If "WTF" folder is found, match it to WoW/WTF</li>
     * <li>If none of the above match, match it to WoW/Interface/Addons/&lt;Name&gt;</li>
     * </ul>
     * Finally, this also updates the MD5 of the addon in question
     * based on what was downloaded
     */
    private File UIReplace(File zip)
        throws /*ZipException,*/ IOException, InterruptedException
    {
/***************************************
 *CONVERTED TO ENCAPSULATOR

        ZipFile file = new ZipFile(zip);
        Enumeration entries = file.entries();
 ****************************************/
        AddonExtractor file = new AddonExtractor(zip);
        file.OpenForReading();
       
        // first, we want to get a list of directories so we can detect
        // our structure
        Vector<File> directories = new Vector<File>();
        String addonsFlag = null, interfaceFlag = null, wtfFlag = null;
        FileReporterHandler.verbose("Z!! "+Thread.currentThread().getName()+" enumerating entries");
        while (file.next())
        {
///////////////////////ZipEntry e = (ZipEntry)entries.nextElement();
            if (file.entryIsDirectory())
            {
                //directories.add(e.getName());
                if (file.entryMatches(".+?Addons/$", true))
                {
                    addonsFlag = file.entryName();
                }
                else if (file.entryMatches(".+?WTF/$", true))
                {
                    wtfFlag = file.entryName();
                }
            }
        }  
        // Now we have our two base paths:
        File addonBase = new File(Main.wowPath, Main.REL_ADDONS);
        File wtfBase = new File(Main.wowPath, "WTF");
        
        if (addonsFlag == null)
            addonsFlag = "#################################";
        if (wtfFlag == null)
            wtfFlag = "#################################";
        
        // Now we want to make a full list of directories in 
        // existance (the directory may not have been listed explicitly,
        // so we must parse the files to find them!)
        file.Close();           // Reopen to reset iterator
        file.OpenForReading();
        
        FileReporterHandler.verbose("Z!! "+Thread.currentThread().getName()+" enumerating directories");
        while (file.next())
        {
            String path = file.entryName();
            
            File f, dir;
            try
            {
                f = MakeFile(addonBase, addonsFlag, wtfBase, wtfFlag, path);
            }
            catch (IllegalArgumentException err)
            {
                continue;       // Skip addons dir
            }
            if (f.isFile())
                dir = f.getParentFile();
            else if (f.exists())
                dir = f;
            else
                continue;       // Skip non-existants
            
            // add directories to list
            directories.add(dir);
        }
        
        FileReporterHandler.verbose("Z!! "+Thread.currentThread().getName()+" destroying old directories");
        // Now go through all the directories, kill them off
        /*
        for (int i = 0; i < directories.size(); i++)
        {
            //@todo I don't know what crack I was smoking when I wrote this, but this is moronic
            // we'll make a billion zip files here from addons like auctioneer. We need to handle
            // this in a single, atomic operation.
            if (directories.get(i).exists())
            {
                // If it exists, then we need to back it up!
                // (Subfolders will have already been deleted)
                File zipBackup = new File(directories.get(i), "backup_"+Main.filetime()+".zip");
                if (zipBackup.exists()) zipBackup.delete();
                zipBackup.createNewFile();
                FileReporterHandler.verbose("Z<< "+Thread.currentThread().getName()+" requests backup of "+directories.get(i).getName());
                CreateBackup(zipBackup, directories.get(i));
                CleanDirectory(directories.get(i), true);      // Removes non-zips from base and deletes subs
            }
        }*/
        BackupAndDelete(directories);
        
        FileReporterHandler.verbose("Z!! "+Thread.currentThread().getName()+" generating files/directories");
        // Now go through the enumeration, make the directories first
////////////////        entries = file.entries();
        file.Close();
        file.OpenForReading();
        UpdateStatus("Installing addon...", 0, 0, file.size(), STANDARD);
        File firstFile = null;
        while (file.next())
        {
//////////////////ZipEntry e = (ZipEntry)entries.nextElement();
            if (file.entryIsDirectory())
            {
                FileReporterHandler.verbose("Z>> "+Thread.currentThread().getName()+" installs directory "+file.entryName());
                String path = file.entryName();
                File f;
                try
                {
                    f = MakeFile(addonBase, addonsFlag, wtfBase, wtfFlag, path);
                }
                catch (IllegalArgumentException err)
                {
                    continue;   // Skip the actual addons dir, etc.
                }
                // Now we have to delete the directory and make a new one
                // (Back up before deleting)
                f.mkdirs();                     // Make if we need to
                
                // There's nothing really to "extract" in a zip directory
                // So just end here (You extract files...)
            }
            else
            {
                FileReporterHandler.verbose("Z>> "+Thread.currentThread().getName()+" installs file "+file.entryName());
                // Unzip this file to the specified path
                String path = file.entryName();
                File f = MakeFile(addonBase, addonsFlag, wtfBase, wtfFlag, path);
                // Make any dirs leading to it if we need to
                f.mkdirs();
                
                // Delete it if it exists
                if (f.exists()) f.delete();
                f.createNewFile();
                
                // Mark it as first file if needed
                if (firstFile == null && 
                        AddonUpdateProcess.isChild(f.getParentFile(), new File(
                            Main.wowPath, Main.REL_ADDONS))
                       )
                {
                    firstFile = f;
                }
                
                // Now extraction!
                /*
                InputStream in = file.getInputStream(e);
                BufferedOutputStream out = new BufferedOutputStream(
                        new FileOutputStream(f)
                );
                
                
                int b;
                while ((b = in.read()) != -1)
                    out.write(b);
                
                out.flush();
                out.close();
                in.close();
                 */
                file.write(f);
                
               
                
                
                
            }
            UpdateStatus(null, -3, -1, -1, STANDARD); // Increment progress bar
            
        }
        file.Close();
        // If this is a new install, zip was in Interface/Addons.
        // Let's move this to the correct place and update info
        if (zip.getParentFile().equals(
                new File(Main.wowPath, Main.REL_ADDONS)
        ))
        {
            // This was a fresh install...
            File f;
            // Work backwards until we get root in addons
            f = firstFile.getParentFile();
            
            // TODO: (Maybe) First check that we aren't already in the addons folder.
            // If we are, we'll skip this one. Things don't get placed there!
            
            
            while (!f.getParentFile().equals(
                    new File(Main.wowPath, Main.REL_ADDONS)))
            {
                f = f.getParentFile();
            }
            f = new File(f, zip.getName());
            
            if (f.exists())     // We have to get rid of the old one first
                f.delete();
            
            if (!zip.renameTo(f))
                throw new IOException("Cannot move ZIP to addon location '"+
                        f.getAbsolutePath()+"' from '"+zip.getAbsolutePath()+"'");
            zip = f;
            
        }
        
        return zip;         // Return back the true location
        
        
    }  
    
    /** Cleans out the directory, removing all files
     * from the base directory, and removing all subdirectories
     * (thus deleting the files inside them as well).<br /><br />
     *
     * This function should always be called with saveZips as true, which
     * will not delete zip files in the base folder. Any subdirectories
     * containing zips will still be deleted.
     * @param f the directory to clean
     * @param saveZips <tt>true</tt> if zips should be skipped, <tt>false</tt>
     * if they should also be deleted
     */
    public static void CleanDirectory(File f, boolean saveZips)
    {
        /* BUGFIX: We no longer delete directories that are
         * outside of the Addon path. This is to prevent problems
         * with complex addons deleting contents of world of warcraft
         * folder or something like that.
         *
         */
        if (f.isDirectory())
        {
            
            File[] forDeletion;
            if (saveZips)
                forDeletion = f.listFiles(new NonZipFilter());
            else
                forDeletion = f.listFiles();
            
            for (int i = 0; i < forDeletion.length; i++)
                CleanDirectory(forDeletion[i], false);
            
            if (isChild(f, new File(Main.wowPath, Main.REL_ADDONS)))
                f.delete();     // Try to delete, it may fail if we saved zips
            
            
        }
        else
        {
            // File, just delete it
            f.delete();
        }
    }
    
    
    
    private File MakeFile(File addonBase, String addonsFlag,
                            File wtfBase, String wtfFlag, String path)
        throws IllegalArgumentException
    {
        File f;
        /*if (path.compareToIgnoreCase(addonsFlag) == 0 ||
            path.compareToIgnoreCase(wtfFlag) == 0 )
        {
            // Don't remake addons/wtf dir, just go next
            throw new IllegalArgumentException("Synonym Detected");
        }*/

        if (path.matches("^"+addonsFlag))
        {
            // Make this entry as a child of addons
            String s = path.substring(addonsFlag.length());
            f = new File(addonBase, s);
            if (f.getParentFile().equals(addonBase))    // Check if direct child
                addon.addDependent(f);                  // Mark dependent
                
        }
        else if (path.matches("^"+wtfFlag))
        {
            // Entry based on WTF
            String s = path.substring(wtfFlag.length());
            f = new File(wtfBase, s);
        }
        else
        {
            f = EstablishParent(addonsFlag, wtfFlag, path, addonBase, wtfBase);
            // Based on addons, use full path
            if (f == null)
                f = new File(addonBase, path);
            
            if (f.getParentFile().equals(addonBase))    // Check if direct child
                addon.addDependent(f);                  // Mark dependent
        }
        
        return f;
    }
    
    /** Scans a zip file and alerts the dependents that they need
     * to be interrupted (this is done by flagging the semaphore of
     * an interrupt state and indicating the newly found dependents)
     * <br /><br />
     * We MUST NOT flag ourselves or we risk interrupting ourselves!
     */
    private void AlertDependents(File f) throws IOException
    {
        if (addon.GetTOC() == null)     // Nothing to alert; new install
            return;
        
//////////ZipFile zf = new ZipFile(f);
        AddonExtractor file = new AddonExtractor(f);
//////////Enumeration entries = zf.entries();
        file.OpenForReading();
        boolean foundDependent = false;
        
        Vector<String> dirsFound = new Vector<String>();
        
        while (file.next())
        {
////////////ZipEntry e = (ZipEntry)entries.nextElement();
            // I know, this looks ugly, let me just explain it, as easily as I can
            // It takes off the last / in the entry name, if it exists.
            // Plain and simple :P
            String s = file.entryName().substring(0,file.entryName().length()-(file.entryMatches("^.+/$", false)?1:0));
            if (file.entryIsDirectory() && 
                    !addon.GetTOC().getParent().toLowerCase().contains(s.toLowerCase()))
            {
                /*
                foundDependent = true;
                Main.f.dependents.add(e.getName());
                 */
                foundDependent = SendAlert(s);
            }
            else
            {
                // File, but we can still put together a list of dirs
                // remember, sometimes zips don't explicitly define dirs
                String n = file.entryName();
                Pattern p = Pattern.compile("Addons/([^/]+)", Pattern.CASE_INSENSITIVE);
                Matcher m = p.matcher(n);
                if (m.find())
                    n = m.group(1);
                // else n = n
                
                // also, we only want the path, so extract it
                // This will also remove any ending backslashes
                n = n.replaceAll("(?:(^.+)/$|(^.+)/[^/]+$|^(.+)$)", "$1$2$3");
                {
                    //n = m.group(1);         // This is the name we want
                    
                    // Same ugly thing as above....
                    if (!addon.GetTOC().getParent().toLowerCase().contains(n.toLowerCase()))
                    {
                        /*
                        foundDependent = true;
                        Main.f.dependents.add(n);
                         */
                        foundDependent = SendAlert(n);
                         
                    }
                }
                
            }
        }
        if (foundDependent)
            Main.f.setFrameState(MainFrame.INTERRUPTED_STATE);
        
        
    }
    
 
    private void CheckInterruption() throws InterruptedException
    {
        if (Thread.currentThread().isInterrupted() || !running)
            throw new InterruptedException();
    }
    
    /** Gets the path to this addon (the parent of the TOC)
     * @return a String representation of the path to this addon
     */
    public String GetPath()
    {
        return addon.GetTOC().getParent();
    }
    
    /** Tests against currently running/pending processes amd
     * adds appropriate items to the alerts semaphore
     * returns true if an alert was added
     */
    private boolean SendAlert(String n)
    {
        for (int i = 0; i < Main.f.runnablesPending.size(); i++)
        {
            if (((AddonUpdateProcess)Main.f.runnablesPending.get(i)).GetPath().toLowerCase().contains(n.toLowerCase()))
            {
                Main.f.dependents.add(n);
                return true;
            }
        }
        return false;
    }
    
    
    private void GetUID()
    {
        UpdateStatus("Suspended: More Information Required", 0, 0, 1, SUSPENDED);
        // We'll remove the progress bar temporarily and place
        // a button there to give more info
        Suspender s = new UIDRequiredSuspender(addon);   
        try
        {
            s.Suspend(this.frameConstraint, Main.f.getComponentPane(), bar);
        }
        catch (InterruptedException e)
        {
            FileReporterHandler.out(">> Caught IRQ-2, terminating thread "+
                    Thread.currentThread().getName());
            running = false;
            return;
        }
        catch (CancellationException e)
        {
            // Delete the queueing file (WID) and mark it
            File wid = new File(addon.toc.getParent(), Main.DATA_FILE);
            if (wid.exists()) 
            {
                wid.delete();
                UpdateStatus("Cancelled - User removed from updater",
                        0,0,1,ERROR);
                throw e;
            }
            else
                throw e;        // Something strange happened; throw it
        }

    }
    
    /** Similar to GetUID(), but checks for a possible detection */
    private void CheckUID(Vector<FavoritesEntry> matches)
    {
        try
        {
            UpdateStatus("Suspended: Clarification Required", 0, 0, 1, SUSPENDED);
            // We have a bunch of matches. Ask which is correct.
            MatcherSuspender s = new MatcherSuspender(matches, addon);
            s.Suspend(this.frameConstraint, Main.f.getComponentPane(), bar);
        }
        catch (InterruptedException e)
        {
            FileReporterHandler.out(">> Caught IRQ-3, terminating thread "+
                    Thread.currentThread().getName());
            running = false;
            return;
        }
            
    }

    /** Backs up anything in the Interface/Addons folder inside 'directories' */
    private void BackupAndDelete(Vector<File> directories) 
        throws IOException, InterruptedException
    {
        // First let's go through and remove duplicates, which may occur,
        // as well as getting rid of anything that (a) doesn't exist, or 
        // (b) is not a directory
        Vector<File> done = new Vector<File>();
        
        for (int i = 0; i < directories.size(); i++)
        {
            boolean duplicate = false;
            for (int j = 0; j < done.size(); j++)
            {
                if (directories.get(i).equals(done.get(j)))
                    duplicate = true;           // Flag as duplicate
            }
            if (!duplicate && isChild(directories.get(i), 
                    new File(Main.wowPath, Main.REL_ADDONS)))
                done.add(directories.get(i));   // Add to list if not duplicate
        }
        
        // Now we have a normalized list; let's do a backup
        File backupsDir = new File(Main.wowPath, Main.REL_BACKUPS);
        if (!backupsDir.exists())
            backupsDir.mkdir();
        
        File file = new File(backupsDir, "backup-"+addon.GetUID()+"-"+Main.filetime()+".zip");
        
        if (!file.exists())
            file.createNewFile();
        else
        {
            file.delete();
            file.createNewFile();
        }
        /*
        FileOutputStream fos = new FileOutputStream(file);
        ZipOutputStream zip = new ZipOutputStream(fos);
         */
        AddonExtractor zip = new AddonExtractor(file);
        zip.OpenForWriting();
        FileReporterHandler.verbose("Z## "+Thread.currentThread().getName()+" ready to backup - "+done.size()+" groups to zip.");
        for (int i = 0; i < done.size(); i++)
        {
            UpdateStatus("Backing up current interface...", i, 0, done.size(), STANDARD);
            AddZipFolder(zip, done.get(i));
            
            // Now do the saved vars
            AddSavedVars(zip, done.get(i));
        }
        
        FileReporterHandler.verbose("Z!! "+Thread.currentThread().getName()+" zip ready; flushing buffers.");
        try
        {
            zip.Close();
        }
        catch (IOException e)
        {
            // Failed to backup because nothing to back up
        }
        
        
        
        
        // Now we can go and delete all this crap
        for (int i = 0; i < done.size(); i++)
            CleanDirectory(done.get(i), true);
        
        // Now add it to the backup manager
        Main.backupManager.OpenForAppending();
        Main.backupManager.AddBackup(file, this.addon, "Installed version "+addon.GetVersion());
        Main.backupManager.Close();
        
        
        
    }

    /** Adds a directory to the specified zip file, as a backup */
    private static void AddZipFolder(AddonExtractor zip, File directory) 
        throws IOException
    {
        FileReporterHandler.verbose("Z<< "+Thread.currentThread().getName()+" begins backing up "+directory.getAbsolutePath());
        File[] list = directory.listFiles(new NonZipFilter());
        for (int i = 0; i < list.length; i++)   // Add each element
        {  
            //FileReporterHandler.verbose("  < "+list[i].getAbsolutePath());
            File element = list[i];
            // Recursively call for directories; get all the files
            
            if (element.isDirectory())
            {
            //    AddZipFolder(zip, element);
                continue;       // Don't bother writing a folder, it doesn't mean
            }                   // anything in a zip anyway
             
            String relPath = getRelativePath(element, Main.wowPath, '/');
/****************************************
            FileReader fr = new FileReader(element);
            ZipEntry ze = new ZipEntry(relPath);
            zip.putNextEntry(ze);
            // Now start writing to the stream for this entry
            int bt;
            while ((bt = fr.read()) != -1)
                zip.write(bt);      // write the data
            
            // Finished writing this element; close it
            zip.closeEntry();
 *****************************************/
            zip.add(element, relPath);
            
        }
    }

    /** This cool little function gives you a string representation of the 
     * path of a file, relative to a parent file, separated by a specified char
     * @param child The child <tt>File</tt> to get a relative path to
     * @param parent The parent <tt>File</tt> to base the path off of
     * @param separator A character to separate the files, typically '/'
     * @return a String representation of a relative path between <tt>child</tt>
     * and <tt>parent</tt>
     */
    public static String getRelativePath(File child, File parent, char separator) 
    {
        String retval = child.getName();
        File currentFile = child.getParentFile();
        do
        {
            retval = currentFile.getName()+separator+retval;
            if (currentFile.equals(parent))
                break;
            
            currentFile = currentFile.getParentFile();
        } while (currentFile != null);
        
        if (currentFile == null)
            return null;            // not a child
        else
            return retval;
    }
    
    /** Works out the file based on parenting information, or null if
     * cannot determine the absolute path
     */
    private static File EstablishParent(String addonsFlag, String wtfFlag, 
            String path, File addonBase, File wtfBase)
    {
        String[] pathList = path.split("/");
        // Try to get a parent file
        File f = GrabParent(addonsFlag, pathList[0], addonBase);
        if (f == null)
        {
            f = GrabParent(wtfFlag, pathList[0], wtfBase);
            if (f == null)
                return null;    // Can't establish a parent
        }
                
        for (int i = 1; i < pathList.length; i++)
            f = new File(f, pathList[i]);
        
        return f;
        
    }
    
    /** Checks to see if the flag is based off the base */
    private static File GrabParent(String flag, String pathBase, File base)
    {
        File f = null;
        if (flag.matches("^"+regescape(pathBase)+".+"))
        {
            // This is a parent; work it out
            // We can do this by the difference of / characters in path
            int afCount = 0;
            int pathCount = 0;
            for (int i = 0; i < flag.length(); i++)
            {
                if (flag.charAt(i) == '/')
                    afCount++;
            }
            for (int i = 0; i < pathBase.length(); i++)
            {
                if (pathBase.charAt(i) == '/')
                    pathCount++;
            }
            // Now we have the difference, let's get parent
            f = base;
            for (int i = 0; i < (afCount - pathCount - 1); i++)
                f = f.getParentFile();
        }
        return f;
    }
    
    /** escapes a regular expression making it a direct match */
    private static String regescape(String s)
    {
        String out = "";
        for (int i = 0; i < s.length(); i++)
        {
            char c = s.charAt(i);
            if (c == '\\' || c == '[' || c == ']' ||
                    c == '+' || c == '-' || c == '?' ||
                    c == '(' || c == ')')
                out += "\\";
            out += c;
        }
        return out;
    }
    
    /** Checks to see if a file is a child of a parent
     * @param file the child <tt>File</tt> to test
     * @param parent the <tt>File</tt> to test the child against
     * @return <tt>true</tt> if <tt>file</tt> is a child of 
     * (and not equivalent to) <tt>parent, false</tt> otherwise
     */
    public static boolean isChild(File file, File parent)
    {
        File thisFile = file.getParentFile();
        while (thisFile != null)
        {
            if (parent.equals(thisFile))
                return true;
            
            thisFile = thisFile.getParentFile();
        };
        return false;
    }

    
    /** backs up the saved vars from the directory specified
     * assumes the TOC is in this directory. If it is not, then it will
     * do nothing.
     */
    private void AddSavedVars(AddonExtractor zip, File directory) 
        throws IOException
    {
        // Optimization. Only do this once... any more is repetative
        if (savedVars == null)
        {
            savedVars = new Vector<File>();
            savedVarsPerChar = new Vector<File>();
            // Build the per-account and per-character vector
            File svLocation = new File(Main.wowPath, Main.REL_SV);
            File[] accounts = svLocation.listFiles();
            for (int i = 0; i < accounts.length; i++)
            {
                // Double check this is actually an account first!
                if (!accounts[i].isDirectory())
                {
                    FileReporterHandler.verbose("Z!! '"+Thread.currentThread().getName()+
                                    "' skipped SV account '"+accounts[i].getAbsolutePath()+
                                    "': Does not exist or not a directory");
                    continue;
                }
                
                // Base SV
                File thisLocation = new File(accounts[i], SV_BASE_LOCATION);
                savedVars.add(thisLocation);
                // Per character SV
                File[] servers = accounts[i].listFiles();
                if (servers == null)
                    continue;
                for (int j = 0; j < servers.length; j++)
                {
                    if (servers[j].equals(thisLocation))
                        continue;   // Skip base SV folder

                    // Got a server, let's find the characters in it
                    File[] chars = servers[j].listFiles();
                    if (chars == null)
                        continue;
                    for (int k = 0; k < chars.length; k++)
                    {
                        // Got a character, now. Grab SV in it and add it
                        // but only if we're sure it exists!!
                        File f = new File(chars[k], SV_BASE_LOCATION);
                        if (f.isDirectory())
                            savedVarsPerChar.add(new File(chars[k], SV_BASE_LOCATION));
                        else
                            FileReporterHandler.verbose("Z!! '"+Thread.currentThread().getName()+
                                    "' skipped SV location '"+f.getAbsolutePath()+"': Does not exist or not a directory");
                            
                    }
                }
            }
        }
        
        // Now we've got all the files we need.
        // Search for any matching...
        
        // name is -1 size because we don't want the trailing /
        String name = directory.getName().substring(0, directory.getName().length());
        
        // Now, go through each SV entry looking for something related
        // to this addon, and add it.
        for (int i = 0; i < savedVars.size(); i++)
        {
            File[] listing = savedVars.get(i).listFiles(new SVFilter(name));
            for (int j = 0; j < listing.length; j++)
            {
                String path = getRelativePath(listing[j], Main.wowPath, '/');
                zip.add(listing[j], path);
            }            
        }
        
        // The same goes for the SVPC entries...
        for (int i = 0; i < savedVarsPerChar.size(); i++)
        {
            File[] listing = savedVarsPerChar.get(i).listFiles(new SVFilter(name));
            if (listing == null)
                JOptionPane.showMessageDialog(null, "File listing null; about to throw exception. ["+savedVarsPerChar.get(i).getAbsolutePath()+"]", "INTERNAL ERROR", JOptionPane.ERROR_MESSAGE);
            for (int j = 0; j < listing.length; j++)
            {
                String path = getRelativePath(listing[j], Main.wowPath, '/');
                zip.add(listing[j], path);
            }            
        }
        
    }
                    
}
/* DEBUG CODE
class ProgressBarTester implements Runnable
{
    JProgressBar bar;
    ProgressBarTester(JProgressBar b)
    {
        bar = b;
    }
    public synchronized void run() {
        //Thread.currentThread().setDaemon(true);
        while (true)
        {
            System.out.println(bar.isVisible()?"Visible":"Invisible");
            try
            {
                this.wait(1000);
            }
            catch (InterruptedException e)
            {
            
            }
        }
    }
 }
 **/
