/*
 * BackupManager.java
 *
 * Manages a list of backups so that the program can revert, add, delete,
 * etc. to a backup list. This can be related to the System Restore for
 * M$ Windows (but a lot, lot better than their pos).
 *
 * It is thread-safe, meaning multiple threads can access the same manager,
 * and this is the recommended method. If the manager is currently open
 * and another thread attempts to open it for writing or reading, it will
 * wait until it has been closed before authorizing the request. It will
 * not return until that request has been granted.
 *
 * Created on April 6, 2007, 4:46 AM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 *
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;

import java.io.*;
import java.util.Calendar;
import java.util.Vector;

/**
 * Manages a list of backups so that the program can revert, add, delete,
 * etc. to a backup list. This can be related to the System Restore for
 * M$ Windows (but a lot, lot better than their pos).
 * <br /><br />
 * It is thread-safe, meaning multiple threads can access the same manager,
 * and this is the recommended method. If the manager is currently open
 * and another thread attempts to open it for writing or reading, it will
 * wait until it has been closed before authorizing the request. It will
 * not return until that request has been granted.
 *
 * @author m.delbuono
 */
public class BackupManager {
    private File f = new File(Main.wowPath, Main.REL_BACKUPLIST);
    /** Status codes for the various modes this manager may be in */
    public static final int
            MODE_CLOSED = 0,
            MODE_READING = 1,
            MODE_WRITING = 2,
            MODE_APPENDING = 3;
    private volatile int mode = MODE_CLOSED;
    private FileReader in = null;
    private FileWriter out = null;
    private BufferedReader inStream = null;
    private BufferedWriter outStream = null;
    private Vector<BackupEntry> entries = new Vector<BackupEntry>();
    private int location = 0;
    
    
    /** Creates a new instance of BackupManager */
    public BackupManager() throws IOException {
        if (!f.exists())        // ensure the file exists at creation
        {
            f.getParentFile().mkdirs(); // Make the folder if needed
            f.createNewFile();          // Then the manifest file
        }
    }
    
    /** Opens the manager for reading. This function will not return
     * until it has opened the manager for reading, which implies that it
     * will wait for the lock to be released, and control the lock at time
     * of return. 
     * @throws IOException if an I/O error occurred during opening
     * @throws InterruptedException if the thread was interrupted while
     * waiting for the lock on the manager to be released
     */
    public synchronized void OpenForReading() 
        throws IOException, InterruptedException
    {
        FileReporterHandler.verbose("C++ "+Thread.currentThread().getName()+
                " enters BackupManager.OpenForReading()");
        while (mode != MODE_CLOSED)
        {
            FileReporterHandler.verbose("W++ "+Thread.currentThread().getName()+
                    " waits for notification on 'this' (BackupManager)");
            this.wait();
            FileReporterHandler.verbose("W-- "+Thread.currentThread().getName()+
                " received notify from 'this' (BackupManager)");
        }
        
        
        in = new FileReader(f);
        inStream = new BufferedReader(in);
        mode = MODE_READING;
        
        FileReporterHandler.verbose("C-- "+Thread.currentThread().getName()+
                " exits BackupManager.OpenForReading()");
    }
    
    /** Opens the manager for writing. This function will not return
     * until it has opened the manager for writing, which implies that it
     * will wait for the lock to be released, and control the lock at time
     * of return.  
     * @throws IOException if an I/O error occurred during opening
     * @throws InterruptedException if the thread was interrupted while
     * waiting for the lock on the manager to be released
     */
    public synchronized void OpenForWriting() 
        throws IOException, InterruptedException
    {
        FileReporterHandler.verbose("C++ "+Thread.currentThread().getName()+
                " enters BackupManager.OpenForWriting()");
        while (mode != MODE_CLOSED)
        {
            FileReporterHandler.verbose("W++ "+Thread.currentThread().getName()+
                    " waits for notification on 'this' (BackupManager)");
            this.wait();
            FileReporterHandler.verbose("W-- "+Thread.currentThread().getName()+
                " received notify from 'this' (BackupManager)");
        }
        
        
        out = new FileWriter(f, false);
        outStream = new BufferedWriter(out);
        mode = MODE_WRITING;
        
        FileReporterHandler.verbose("C-- "+Thread.currentThread().getName()+
                " exits BackupManager.OpenForWriting()");
    }
    
    /** Opens the manager for appending. This function will not return
     * until it has opened the manager for appending, which implies that it
     * will wait for the lock to be released, and control the lock at time
     * of return. 
     * @throws IOException if an I/O error occurred during opening
     * @throws InterruptedException if the thread was interrupted while
     * waiting for the lock on the manager to be released
     */
    public synchronized void OpenForAppending() 
        throws IOException, InterruptedException
    {
        FileReporterHandler.verbose("C++ "+Thread.currentThread().getName()+
                " enters BackupManager.OpenForAppending()");
        while (mode != MODE_CLOSED)
        {
            FileReporterHandler.verbose("W++ "+Thread.currentThread().getName()+
                    " waits for notification on 'this' (BackupManager)");
            this.wait();
            FileReporterHandler.verbose("W-- "+Thread.currentThread().getName()+
                " received notify from 'this' (BackupManager)");
        }
        
        
        out = new FileWriter(f, true);
        outStream = new BufferedWriter(out);
        mode = MODE_APPENDING;
        
        FileReporterHandler.verbose("C-- "+Thread.currentThread().getName()+
                " exits BackupManager.OpenForAppending()");
    }
    
    /** Closes any streams and sets mode to closed, then releases locks
     * @throws IOException if an error occurred while closing
     */
    public synchronized void Close() throws IOException
    {
        if (inStream != null)
        {
            inStream.close();
            inStream = null;
        }
        if (outStream != null)
        {
            outStream.close();
            outStream = null;
        }
        if (in != null)
        {
            in.close();
            in = null;
        }
        if (out != null)
        {
            out.close();
            out = null;
        }
        
        entries.clear();
        
        mode = MODE_CLOSED;
        FileReporterHandler.verbose("W## "+Thread.currentThread()+" sends "+
                "notify to 'this' (BackupManager)");
        this.notify();  // Notify any threads which may be waiting
    }
    
    /** Appends a backup to the list
     * (Requires appending mode)
     * @param location a <tt>File</tt> identifying the location of the backup
     * @param addon an <tt>Addon</tt> identifying the addon this backup
     * relates to, or <tt>null</tt> if it relates to no addon
     * @param reason the reason for the backup
     * @throws IOException if an I/O error occurred or if the backup manager
     * is not in appending mode
     * @see wowiupdater.BackupManager#OpenForAppending()
     */
    public void AddBackup(File location, Addon addon, String reason) 
        throws IOException
    {
        // Verify we're in appending mode
        if (mode != MODE_APPENDING)
            throw new IOException("Backup manager not in appending mode");
        
        // Make a new line
        outStream.newLine();
        // First line is always the timestamp
        Calendar currentTime = Calendar.getInstance();
        long timestamp = currentTime.getTime().getTime();
        outStream.write(String.valueOf(timestamp));
        outStream.newLine();
        // Next line is backup location
        outStream.write(location.getAbsolutePath());
        outStream.newLine();
        // Next comes the UID followed by a NL then the addon name
        // UID is 0 if no addon (null), followed by nothing
        if (addon == null)
        {
            outStream.write("0");
        }
        else
        {
            outStream.write(String.valueOf(addon.GetUID()));
            outStream.newLine();
            outStream.write(addon.GetName());
        }
        outStream.newLine();
        // Finally comes the reasoning (may include version number, etc.)
        outStream.write(reason);
    }
    
    /** Loads an array of backup entries, which contain addon info, 
     * and other such things. Requires reading mode.
     * @throws IOException if an error occurred during reading or if the 
     * backup manager is not in reading mode.
     * @see wowiupdater.BackupManager#OpenForReading()
     */
    public void load() throws IOException
    {
        if (mode != MODE_READING)
            throw new IOException("Backup manager not in reading mode");
        
        String data;
        entries.clear();
        while((data = inStream.readLine()) != null) // While more lines
        {
            // First line is timestamp
            long timestamp;
            try
            {
                timestamp = Long.parseLong(data);
            }
            catch (NumberFormatException e)
            {
                // Just continue onto the next line.
                continue;
            }
            // Next line is location of backup
            File location = new File(inStream.readLine());
            // Next line is UID or zero
            int uid = Integer.parseInt(inStream.readLine());
            // Next line is addon name if and only if UID is not 0
            String name = null;
            if (uid != 0)
                name = inStream.readLine();
            // Next line is reasoning
            String reasoning = inStream.readLine();
            
            // Now make the entry
            entries.add(new BackupEntry(timestamp, location,
                    uid, name, reasoning));
   
        }
        reset();
        
    }
    
    /** Gets the next entry available 
     * @return the next <tt>BackupEntry</tt>
     */
    public BackupEntry next()
    {
        if (location >= entries.size())
            return null;
        
        return entries.get(location++); // get entry, then increment by 1
    }
    
    /** Resets the iterator that is getting next() elements
     * @see wowiupdater.BackupManager#next()
     */
    public void reset()
    {
        location = 0;
    }
    
    
    /** Shows a dialog to the user requesting for revert info
     * (this is used when the user asks to revert to a backup)
     * Returns an array of Files which are to be extracted.
     * @throws UnsupportedOperationException always
     * @deprecated Use wowiupdater.ReversionDialog directly
     * @see wowiupdater.ReversionDialog
     */
    @Deprecated
    public File[] revert()
    {        
        throw new UnsupportedOperationException("Backup reversion is NLI");
    }
    
}
