/*
 * FavoritesEntry.java
 *
 * Abstracts an entry in a favorites list, identifying addons
 * along with their WoWInterface UID.
 *
 * Created on March 3, 2007, 7:37 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;

import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Vector;

/**
 * Abstracts an entry in a favorites list, identifying addons
 * along with their WoWInterface UID.
 * @author m.delbuono
 */
public class FavoritesEntry {
    /** The author of this addon favorite */
    public String author;
    /** The UID representing this addon favorite */
    public int uid;
    /** The name of this addon favorite */
    public String name;
    /** The score of this addon. Used when attempting to judge matches */
    public int score;
    /** The version string of this addon favorite */
    public String version;
    
    private static File[] list; // List of addon folders
    //public boolean missing;// is the favorite entry not found? (Checks WIDs)
    /** Creates a new instance of FavoritesEntry */
    public FavoritesEntry() {
        author = "";
        uid = -1;
        name = "";
        score = 0;
        version = "";
        //missing = true;
    }
    
    /** Represents this addon favorite as a string, for use in lists
     * @return a String representation of this addon favorite
     */
    public String toString() {
        return name + " ("+score+")";
    }
    
    /** Determines if the favorite is currently installed
     * @return <tt>True</tt> if the item was found on the user's folder, 
     * <tt>false</tt> if not
     * @throws IOException if an I/O error occurred
     */
    public boolean isFound() throws IOException
    {
        //WIDFilter filter = new WIDFilter();
        
        // This is an optimization. It only needs to happen once, so
        // why not only do it once. Every other time it would be the same...
        if (list == null)
            list = getAllFiles();
        
        // Now open each one and read the first line. It should match the UID
        // This can be done by constructing the addon from file
        for (int i = 0; i < list.length; i++)
        {
            try
            {
                Addon a = new Addon(list[i].getParent());
                if (a.GetUID() == uid)
                    return true;
            }
            catch (IllegalArgumentException e)
            {
                // Dependent addon. We'll skip this.
                continue;
            }
            catch (IOException e)
            {
                // Ignore these cases too
                continue;
            }
            
        }
        
        return false;
    }

    /** Converts this addon favorite into an <tt>Addon</tt> abstraction
     * @return a new <tt>Addon</tt>
     * @see wowiupdater.Addon
     */
    public Addon toAddon() {
        return new Addon(this);
    }
    
    
    private File[] getAllFiles() throws IOException
    {
        Vector<File> retval = new Vector<File>();
        
        File addons = new File(Main.wowPath, Main.REL_ADDONS);
        if (!addons.exists())
            throw new FileNotFoundException("Unable to isolate addons directory");
        
        // First list out all dirs
        WIDFilter dirFilter = new WIDFilter(true);
        WIDFilter fileFilter = new WIDFilter(false);
        File[] dirs = addons.listFiles(dirFilter);
        
        // Now go through them, recurse ONLY ONCE if necessary
        for (int i = 0; i < dirs.length; i++)
        {
            if (dirs[i].isDirectory())
            {
                File[] dirs2 = dirs[i].listFiles(fileFilter);
                for (int j = 0; j < dirs2.length; j++)
                {
                    if (dirs2[j].isFile())
                        retval.add(dirs2[j]);
                }
            }
            else    // Must be WID file
                retval.add(dirs[i]);
        }
        
        File[] ret = new File[retval.size()];
        for (int i = 0; i < retval.size(); i++)
        {
            ret[i] = retval.get(i);
        }
        return ret;
    }
    
}
