/*
 * FavoritesList.java
 *
 * Stores a list of favorites, which is accessible via the web using
 * XML/RSS feeds, and allows WoWIUpdater to identify addons based
 * on their UID.
 *
 * Created on March 3, 2007, 7:35 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;
import java.util.Vector;
import java.util.regex.*;
import java.io.*;
/**
 * Stores a list of favorites, which is accessible via the web using
 * XML/RSS feeds, and allows WoWIUpdater to identify addons based
 * on their UID.
 * @author m.delbuono
 */
public class FavoritesList {
    private Vector<FavoritesEntry> entries;
    /** Creates a new instance of FavoritesList */
    public FavoritesList() {
        entries = new Vector<FavoritesEntry>();
    }
    
    /** Loads a favorites list from a WIC file
     * Files should be written in the following form:
     * Each line in a favorites file should appear as
     * <br />UID:Name
     * <br />UID:Name
     * <br />...
     * @param f the File to load
     * @throws FileNotFoundException if the file does not exist
     * @throws IOException if an I/O error occurred
     */
    public void load(File f) throws FileNotFoundException, IOException
    {
        if (!f.exists())
            throw new FileNotFoundException("No favorites file available");
        
        String l;
        FileReader fr = new FileReader(f);
        BufferedReader reader = new BufferedReader(fr);
        Pattern p = Pattern.compile(
                "([0-9]+):([^:]+):(.+)", 
                Pattern.CANON_EQ
        );
        while ((l = reader.readLine())!=null)
        {
            Matcher m = p.matcher(l);
            if (!m.find())
                throw new IOException("Bad favorites file");
            
            // We have a match, let's extract the UID, author, and name
            FavoritesEntry e = new FavoritesEntry();
            e.uid = Integer.parseInt(m.group(1));
            e.author = m.group(2);
            e.name = m.group(3);
            
            // Add it
            entries.add(e);
        }
        
    }
    
    /* Attempts to identify a favorites entry by addon folder name by
     * using a complex point-system algorithm.
     * @param addon the name of the addon
     * @param author the name of the author, or null if unknown
     * @return Vector<FavoritesEntry> a sorted list of closest matching entries, descending, or null if all entries gave 0 points
     */
    public Vector<FavoritesEntry> match(String addon, String author)
    {
        if (author == null)
            author = "";
        
        Vector<FavoritesEntry> entries; // the working entries for this match
        entries = new Vector<FavoritesEntry>(this.entries);// copy it
        
        /* Matching algorithm:
         *   Convert all underscores to spaces
         *   Remove all non-alphanumeric characters
         *   Search for converted pathname inside fav name, word by word
         *      5 points for every matching word
         *   Give 10 points for being the shortest match with >= 5 points
         *       (10-x) points for being not shortest match, where
         *               x is the number of characters longer than
         *               the shortest match. No negative numbers
         *   10 points for having the TOC author found in fav author
         */

        // Step 1: All underscores become spaces
        String convertedName = addon;
        convertedName.replace('_',' ');

        // Step 2: All non-alphanumerics are removed
        convertedName = 
                Pattern.compile("[^a-zA-Z0-9]").matcher(convertedName).replaceAll("");
        
        // Step 2.5: Split the string
        Vector<String> searchers = StringArrayToVector(convertedName.split(" "));
        Vector<String> authors = StringArrayToVector(author.split(" "));
        
        // Step 2.75: Only 4+ chars are matchable
        for (int i = 0; i < searchers.size(); i++)
            if (searchers.get(i).length() < 4) searchers.remove(i);
        for (int i = 0; i < authors.size(); i++)
            if (authors.get(i).length() < 4) authors.remove(i);
        
        // Step 3: Search inside favorites name
        int shortest = 1000;                // Some massive number to start
        for (int i = 0; i < entries.size(); i++)
        {
            entries.get(i).score = 0;      // Reset each entry's score to start
            for (int j = 0; j < searchers.size(); j++)
            {
                if (entries.get(i).name.contains(searchers.get(j)))
                {
                    entries.get(i).score += 5; 
                    if (shortest > entries.get(i).name.length())
                        shortest = entries.get(i).name.length();
                }
            }
        }
        
        // Step 4: Give points for most concise answers
        for (int i = 0; i < entries.size(); i++)
        {
            if (entries.get(i).score > 0 &&   // Skip non-matching entries
                entries.get(i).name.length()-shortest < 10) // and too long ones
            {
                entries.get(i).score +=
                        10-(entries.get(i).name.length()-shortest);
            }            
        }
        
        // Step 5: Give 10 points for a matched author
        for (int i = 0; i < authors.size(); i++)
        {
            for (int j = 0; j < entries.size(); j++)
            {
                if (entries.get(j).author.toLowerCase().contains(authors.get(i).toLowerCase()))
                    entries.get(j).score += 10;
            }
        }
        
        // Step 6: Sort by score
        // Using a reversed search sort to handle this
        for (int i = 0; i < entries.size(); i++)
        {
            for (int j = i; j < entries.size(); j++)
            {
                if (entries.get(j).score > entries.get(i).score)
                {
                    // Push to front
                    entries.insertElementAt(entries.remove(j), i);
                }
            }
        }
        
        // Step 7: Remove all 0-score entries
        for (int i = 0; i < entries.size(); i++)
        {
            if (entries.get(i).score <= 0)
            {
                entries.remove(i);
                i--;                // Compensate for the loss of an element
            }
        }
        
        // finally, return our matches!
        return entries;
        
    }
    
    /** Convenience function to convert string array to vector of strings 
     * @param s an array of strings
     * @return a vector of strings matching the array of strings
     */
    public static Vector<String> StringArrayToVector(String[] s)
    {
        if (s == null)
            return new Vector<String>();        // Return an empty vector
        
        Vector<String> v = new Vector<String>();
        for (int i = 0; i < s.length; i++)
            if (s[i] != "" && s[i] != null) v.add(s[i]);
        
        return v;
    }
    /** Adds an entry to this favorites list
     * @param e the <tt>FavoritesEntry</tt> to add
     */
    public void addEntry(FavoritesEntry e)
    {
        entries.add(e);
    }
    
    /** Determines the size of this list
     * @return the number of elements in the list
     */
    public int size()
    {
        return entries.size();
    }
    
    /** Gets a given entry in the list
     * @param i the zero-based index of the entry to retrieve
     * @return the requested <tt>FavoritesEntry</tt>
     */
    public FavoritesEntry get(int i)
    {
        return entries.get(i);
    }
    
    /** Appends a list of entries to the favorites list
     * @param l the list of entries to add
     */
    public void addEntry(FavoritesList l)
    {
        for (int i = 0; i < l.size(); i++)
        {
            this.addEntry(l.get(i));
        }
    }
    
    /** Determines if the given favorite is installed
     * @param entry the zero-based index of the entry to test
     * @return <tt>true</tt> if the entry is installed, <tt>false</tt> if not
     * @throws IOException if an I/O error occurred
     */
    public boolean isFound(int entry) throws IOException
    {
        return get(entry).isFound();
    }
}
