/*
 * FavoritesParser.java
 *
 * Parses an XML list of favorites, which is accessible via the web using
 * XML/RSS feeds, and allows WoWIUpdater to identify addons based
 * on their UID.
 *
 * Created on March 10, 2007, 10:35 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

/**
 * Parses an XML list of favorites, which is accessible via the web using
 * XML/RSS feeds, and allows WoWIUpdater to identify addons based
 * on their UID.
 * @author m.delbuono
 */
public class FavoritesParser extends DefaultHandler {
    private FavoritesList list;
    private FavoritesEntry currentEntry;
    private String buffer;
    
    public static final int
            ELEMENT_NONE = 0,
            //ELEMENT_Favorites = 1,
            //ELEMENT_Ui = 2,
            ELEMENT_UID = 3,
            ELEMENT_UIName = 4,
            ELEMENT_UIAuthorName = 5,
            ELEMENT_UIVersion = 6,
            
            ELEMENT_ErrorID = 100,
            ELEMENT_ErrorMsg = 101;
    
    private int Element = ELEMENT_NONE;
    
    private int errorId = 0;
    private String errMsg = null;
    private boolean inError = false;
    
    /** Creates a new instance of FavoritesParser */
    public FavoritesParser() {
        list = new FavoritesList();     // Make a new list to store info
        currentEntry = null;
    }
    
    public FavoritesParser(FavoritesList l)
    {
        list = l;
        currentEntry = null;
    }
    /* Debug code only
    public void endDocument()
    {
        System.out.println("End!");
    }
    */
    
    public void startElement(   String namespaceURI,
                                String sName,
                                String qName,
                                Attributes attrs
                            )
         throws SAXException
    {
        
        String eName = sName;
        if (eName.equals("")) eName = qName;    // not namespace-aware
        //System.out.println("Start Element "+eName);
        if (eName.equals("Favorites"))
        {
            // We don't need to do anything here, really
        }
        else if (eName.equals("Ui"))
        {
            // Make a new list element
            currentEntry = new FavoritesEntry();
        }
        else if (eName.equals("UID"))
        {
            // Add the UID to the current entry
            Element = ELEMENT_UID;
        }
        else if (eName.equals("UIName"))
        {
            Element = ELEMENT_UIName;
        }
        else if (eName.equals("UIAuthorName"))
        {
            Element = ELEMENT_UIAuthorName;
        }
        else if (eName.equals("UIVersion"))
        {
            Element = ELEMENT_UIVersion;
        }
        else if (eName.equals("error"))
        {
            inError = true;
        }
        else if (eName.equals("id") && inError)
        {
            Element = ELEMENT_ErrorID;
        }
        else if (eName.equals("message") && inError)
        {
            Element = ELEMENT_ErrorMsg;
        }
        else
        {
            throw new SAXException("Unrecognized tag \""+eName+"\"");
        }
        buffer = "";
    }
    
    public void endElement( String namespaceURI,
                            String sName,
                            String qName
                          )
         throws SAXException
    {
        String eName = sName;
        if (eName.equals("")) eName = qName;    // not namespace-aware
        //System.out.println("End Element "+eName);
        if (eName.equals("Favorites"))
        {
            // Do nothing
        }
        else if (eName.equals("Ui"))
        {
            // Close out this entry and add it to the list
            list.addEntry(currentEntry);
            currentEntry = null;
        }
        else if (   eName.equals("UID") ||
                    eName.equals("UIName") ||
                    eName.equals("UIAuthorName") ||
                    eName.equals("UIVersion")
                )
        {
            // We need to commit the changes
            if (Element == ELEMENT_UID)
            {
                int id = Integer.parseInt(buffer);
                currentEntry.uid = id;
            }
            else if (Element == ELEMENT_UIName)
            {
                currentEntry.name = buffer;
            }
            else if (Element == ELEMENT_UIAuthorName)
            {
                currentEntry.author = buffer;
            }
            else if (Element == ELEMENT_UIVersion)
            {
                currentEntry.version = buffer;
            }
        }
        else if (eName.equals("error")) 
        {
            inError = false;
        }
        else if (eName.equals("id") && inError)
        {
            errorId = Integer.parseInt(buffer);
        }
        else if (eName.equals("message") && inError)
        {
            errMsg = buffer;
        }
        else
        {
            throw new SAXException("Unrecognized closing tag "+eName);
        }       
        buffer = "";
    }
    
    public void characters(char buf[], int offset, int len)
        throws SAXException
    {
        String s = new String(buf, offset, len);
        //System.out.println("Get Characters "+s);
        buffer += s;
    }
    
    public FavoritesList getList()
    {
        return list;
    }
    
    public boolean hasError()
    {
        return errorId != 0;
    }
    
    public int getErrorID()
    {
        return errorId;
    }
    
    public String getError()
    {
        return errMsg;
    }
    
}
