/*
 * FileReporterHandler.java
 *
 * Implements a default UncaughtExceptionHandler. This class makes
 * run-time more user-friendly by suppressing any thread-terminating
 * (unhandled) errors, and reporting them to an errors log instead of
 * the System.err stream. The error log can later be accessed if needed
 * or sent from the user to a developer so that debugging can take place.
 * The error log includes a full stack trace, where possible, along with
 * the thread name and error class/message.
 *
 * Created on March 5, 2007, 6:25 AM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.PrintWriter;

/**
 * Implements a default UncaughtExceptionHandler. This class makes
 * run-time more user-friendly by suppressing any thread-terminating
 * (unhandled) errors, and reporting them to an errors log instead of
 * the System.err stream. The error log can later be accessed if needed
 * or sent from the user to a developer so that debugging can take place.
 * The error log includes a full stack trace, where possible, along with
 * the thread name and error class/message.
 * <br /><br />
 * Reports errors to a file with stack trace. Also implements a thread to
 * act as a shutdown hook <b>WHICH MUST BE CALLED TO FLUSH THE BUFFER</b>.
 *
 * @author m.delbuono
 */
public class FileReporterHandler extends java.lang.Thread
        implements java.lang.Thread.UncaughtExceptionHandler
                                 
{
    private static final File f = new File(Main.WD_PATH+"/error.log"), 
            vf = new File(Main.WD_PATH+"/verbose.log");
    private static FileWriter writer;
    //private static BufferedWriter ostream;
    private static volatile boolean beenInit = false;
    /** Creates a new instance of FileReporterHandler */
    public FileReporterHandler() {
        
    }
    
    /** Called once at start of program, this sets up the handler */
    public static synchronized void init()
    {
        if (beenInit)
            return;         // Don't re-init
     
        try
        {
            beenInit = true;

            if (!Main.verboseMode)
            {
                writer = new FileWriter(f, true);
            }
            else
            {
                //System.out.println("In verbose mode!");
                writer = new FileWriter(vf, true);
                
                FileWriter TempWriter = new FileWriter(f, true);
                BufferedWriter TempBW = new BufferedWriter(TempWriter);
                TempBW.newLine();
                TempBW.write(Main.filetime());
                TempBW.newLine();
                TempBW.write(">>> WARNING: User running in verbose mode. See verbose log.");
                TempBW.newLine();
                TempBW.flush();
                TempWriter.flush();
                TempBW.close();
                TempWriter.close();
            }
            //ostream = new BufferedWriter(writer);
            
            if (Main.verboseMode)
                verbose("@@@ System Init");
            
            writer.flush();
        }
        catch (Throwable err)
        {
            System.err.println("System Panic - Unable to initialize error handler");
            err.printStackTrace(System.err);
            System.err.println("- - - Sending Application Halt - - -");
            Runtime.getRuntime().halt(9);
        }
    }
    
    /** Writes a standard output message to the log file 
     * @param s the String to output
     */
    public static synchronized void out(String s)
    {
        try
        {
            writer.write(Main.filetime()+"> "+s);
            nl();
            writer.flush();
        }
        catch(Throwable err)
        {
            System.err.println("System Panic - Unable to write to log");
            err.printStackTrace(System.err);
            System.err.println("- - - Sending Application Halt - - -");
            Runtime.getRuntime().halt(8);
        }
    }
    
    /** Writes an error message to the log file (with an error prefix)
     * @param s the string to write
     */
    public static synchronized void err(String s)
    {
        try
        {
            writer.write("!!! "+Main.filetime()+"> "+s);
            nl();
            writer.flush();
        }
        catch(Throwable err)
        {
            System.err.println("System Panic - Unable to write to log");
            err.printStackTrace(System.err);
            System.err.println("- - - Sending Application Halt - - -");
            Runtime.getRuntime().halt(8);
        }
    }
    
    /** Dumps a <tt>Throwable</tt> to the log file
     * @param t the <tt>Throwable</tt> to dump, with stack trace
     */
    public static synchronized void dump(Throwable t)
    {
        try
        {
            PrintWriter w = new PrintWriter(writer);
            t.printStackTrace(w);
            w.flush();
            nl(); nl();
            writer.flush();
        }
        catch (Throwable err)
        {
            System.err.println("System Panic - Unable to dump to log");
            err.printStackTrace(System.err);
            System.err.println("- - - Sending Application Halt - - -");
            Runtime.getRuntime().halt(8);
        }
    }
    
    /**
     *  Verbose messages help in resolving deadlocks and other strange
     *  issues by dumping a lot of state information to a file. Please note
     *  that this file can become huge and also degrades performance significantly
     *  so it is normally disabled. You can activate it using --verbose on the 
     *  command line, but it is not recommended unless absolutely necessary.
     * <br /><br />
     * <pre>
     * Standard format for verbose strings:
     *      X## - T <string>
     *
     *      X## = Action Code:
     *              L++     Thread Lock Gained
     *              L--     Thread Lock Released
     *              L??     Thread Lock Requested
     *              C++     Function Called
     *              C--     Function Exit
     *              W++     Entering wait state
     *              W--     Leaving wait state
     *              W!!     Broke from wait state via interrupt
     *              W##     Sending interrupt
     *              W**     Sending notify
     *              Z!!     Ziptype initialization
     *              E>>     Encryption information
     *              @@@     FileReporterHandler message
     * 
     *      T = Thread Name
     *      <string> = Message
     * </pre>
     * @param s the verbose message to write
     */
    public static synchronized void verbose(String s)
    {
        try
        {
            if (Main.verboseMode)
            {
                writer.write(Main.filetime()+"> "+s);
                writer.write(" --> ");
                try
                {
                    writer.write(Thread.currentThread().getStackTrace()[3].toString());
                }
                catch (Exception e)
                {
                    try
                    {
                        writer.write(Thread.currentThread().getStackTrace()[2].toString());
                    }
                    catch (Exception err)
                    {
                        writer.write(Thread.currentThread().getStackTrace()[1].toString());
                    }
                }
                
                nl();
                writer.flush();
            }
        }
        catch (java.io.IOException err)
        {
            System.err.println("System Panic - Unable to write to log: "+s);
            err.printStackTrace(System.err);
            System.err.println("- - - Sending Application Halt - - -");
            Runtime.getRuntime().halt(8);
        }
    }

    /** An uncaught exception handler to be used when the thread encounters
     * an uncaught exception. This function will dump an error message
     * and shut down that thread. If an exception occurrs in this function,
     * the JVM will be shut down.
     * @param t the <tt>Thread</tt> where the error occurred
     * @param e the <tt>Throwable</tt> that triggered the exception
     */
    public synchronized void uncaughtException(Thread t, Throwable e) {
        try
        {
            nl(); nl();
            writer.write("**********************************************");
            nl();
            writer.write("*** "+Main.filetime()+" TRAP > "+t.getId());
            nl();
//@todo remove debug code
e.printStackTrace();
            StackTraceElement[] stack = e.getStackTrace();
            
            writer.write("Exception ["+Main.filetime()+
                    "] in thread \""+t.getName()+"\": "+e.getClass().getName()+" ==> "+e.getMessage());
            for (int i = 0; i < stack.length; i++)
            {
                nl();
                writer.write("\tat "+stack[i].getClassName()+"."+
                        stack[i].getMethodName()+"("+
                        stack[i].getFileName()+":"+
                        stack[i].getLineNumber()+") ");
                
            }
            nl();
            writer.write("**********************************************");
            nl(); nl();
            writer.flush();
            
            Main.f.SetStatus("An internal error has been logged. If problems persist, please file a bug report.");
        }
        catch (Throwable err)
        {
            System.err.println("System Panic - Unable to log thread exception");
            err.printStackTrace(System.err);
            System.err.println("Cause >");
            e.printStackTrace(System.err);
            System.err.println("- - - Sending application halt - - -");
            Runtime.getRuntime().halt(8);
        }
        
        
    }

    /** Executes the shutdown hook (cleanup and flush) */
    public void run() {
         try
         {
            verbose("@@@ System Shutdown");
            
            nl(); nl();
            
            //ostream.flush();
            writer.flush();
            //ostream.close();
            writer.close();
        }
        catch (java.io.IOException e)
        {
            System.err.println("Error finalizing shutdown:");
            e.printStackTrace(System.err);
        }
    }
    
    /** Writes a new line to the file 
     * @throws IOException if an I/O error occurs
     */
    private static void nl() throws java.io.IOException
    {
        writer.write(System.getProperty("line.separator"));
    }
}
