/*
 * ResolverSuspender.java
 *
 * Implements a combination of a Suspender and ActionListener to
 * abstract a case of a Suspender where a frame will be shown, and the
 * program will wait for user input. One possible case of this is where a 
 * user must resolve a conflicting condition where the program cannot
 * detect the UID of an addon (which is a direct extension of this class,
 * UIDRequiredSuspender, at UIDRequiredSuspender.java)
 *
 * Created on March 4, 2007, 5:21 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;

/**
 * Implements a combination of a Suspender and ActionListener to
 * abstract a case of a Suspender where a frame will be shown, and the
 * program will wait for user input. One possible case of this is where a 
 * user must resolve a conflicting condition where the program cannot
 * detect the UID of an addon (which is a direct extension of this class,
 * UIDRequiredSuspender, at UIDRequiredSuspender.java)
 *
 * @author m.delbuono
 */
public abstract class ResolverSuspender implements Suspender, ActionListener {
    
    /** Creates a new instance of ResolverSuspender */
    /*public ResolverSuspender() {
    }*/
    
    /** The <tt>JButton</tt> that the user may click to resolve this suspension */
    protected JButton resolveBtn;
    /** The <tt>JProgressBar</tt> indicating status for this addon */
    protected JProgressBar tempBar, oldBar;    
    /** The <tt>JPanel</tt> where this <tt>Suspender</tt>'s controls may fit */
    protected JPanel p;
    /** The <tt>GridBagConstraints</tt> controlling this <tt>Suspender</tt> */
    protected GridBagConstraints constraints;
         
    /** While resolving, we don't want them to be able to open another
     * resolving thread, or bad things happen (yes, I tried, not pretty :P)
     * We'll replace the resolve button with an indeterminate progress bar
     */
    protected void DisableResolve()
    {        
        resolveBtn.setVisible(false);
        
        if (tempBar == null)
            tempBar = new JProgressBar();
        
        p.add(tempBar, constraints);
        tempBar.setVisible(true);
        tempBar.setIndeterminate(true);
        p.validate();
    }
    
    /** Re-enables resolution which was disabled by <tt>DisableResolve()</tt>
     * @see wowiupdater.ResolverSuspender#DisableResolve()
     */
    protected void EnableResolve()
    {
        tempBar.setVisible(false);
        resolveBtn.setVisible(true);
        p.remove(tempBar);
        p.validate();
    }
    
    /** Creates a resolution button that the user can click on to resolve
     * this suspension
     * @param text the text to be displayed on the button
     */
    protected void GenerateResolutionButton(String text)
    {
        Main.suspendedCount++;
        oldBar.setVisible(false);
        resolveBtn = new JButton(text);
        resolveBtn.addActionListener(this);
        
        p.add(resolveBtn, constraints);
        p.validate();
        
    }
    
    /** Disposes the resolution button created by GenerateResolutionButton()
     * @see wowiupdater.ResolverSuspender#GenerateResolutionButton(String text)
     */
    protected void DisposeResolutionButton()
    {
        Main.suspendedCount--;
        p.remove(resolveBtn);
        oldBar.setVisible(true);
    }
    
    /** Checks to see if the current thread is interrupted (meaning it 
     * should quit because <tt>MainFrame</tt> asked for a rebuild)
     * @throws InterruptedException if the current thread is interrupted
     * @see wowiupdater.Main#rebuild()
     */
    protected void CheckForRebuild() throws InterruptedException
    {
        if (Thread.currentThread().interrupted())
            throw new InterruptedException();
    }
    
}
