/*
 * UIDRequiredSuspender.java
 *
 * Extends a ResolverSuspender which requests that the user resolve a
 * conflicted state because the program cannot detect the addon's UID
 * automatically.
 *
 * Created on March 3, 2007, 11:28 PM
 *
 * Copyright (c) 2007 Matthew P. Del Buono
 * 
 * 
 *     This file is part of the WoWInterface UI Manager (WoWIUpdater)
 *
 *     WoWIUpdater is free software; you can redistribute it and/or modify 
 *     it under the terms of the GNU General Public License as published by 
 *     the Free Software Foundation; either version 2 of the License, or 
 *     (at your option) any later version.
 *
 *     WoWIUpdater is distributed in the hope that it will be useful, but 
 *     WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *     General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 *     To contact the author, please send an e-mail to
 *     Matthew.Delbuono@erau.edu or contact "Shirik" via forum private
 *     message at http://www.wowinterface.com or contact the following
 *     physical address:
 *
 *          Matthew P. Del Buono
 *          600 S Clyde Morris Blvd
 *          Mail #146379
 *          Daytona Beach, FL 32114
 *          United States of America
 */

package wowiupdater;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
/**
 * Extends a <tt>ResolverSuspender</tt> which requests that the user resolve a
 * conflicted state because the program cannot detect the addon's UID
 * automatically.
 *
 * @author m.delbuono
 */
public class UIDRequiredSuspender extends ResolverSuspender {
    private Addon addon;
    private JButton btnOK, btnCancel, btnDelete;
    private UIDFrame frame;
    //boolean isResolved = false;
    private volatile JButton resolution = null;
    
    
    /** Creates a new instance of UIDRequiredSuspender */
    public UIDRequiredSuspender(Addon a) {
        addon = a;
    }
    
    /** Requests that the current state be suspended until a UID can be 
     * supplied.
     * @param c the <tt>GridBagConstraints</tt> binding this <tt>Suspender</tt>
     * @param p the <tt>JPanel</tt> where this <tt>Suspender</tt> may display information
     * @param b the <tt>JProgressBar</tt> that shows this addon's status
     * @throws java.util.concurrent.CancellationException if the operation was
     * cancelled
     * @throws InterruptedException if the thread entered an interrupted state
     * while in a wait lock
     */
    public synchronized void Suspend(GridBagConstraints c, JPanel p, JProgressBar b)
        throws java.util.concurrent.CancellationException, InterruptedException
    {
        FileReporterHandler.verbose("C++ "+Thread.currentThread().getName()+" enters UIDRequiredSuspender.Suspend");
        //isResolved = false;
        resolution = null;
        constraints = c;
        
        
        // Replace progress bar with a button
        this.p = p;
        oldBar = b;
        GenerateResolutionButton("Resolve");
        
                
        while (resolution == null      // via another thread
               || resolution == btnCancel)
        {
            resolution = null;          // in case cancel was set
            Main.f.repaint();
            
            try
            {
                FileReporterHandler.verbose("W++ "+Thread.currentThread().getName()+" waits for notification from 'this'");
                this.wait();
                FileReporterHandler.verbose("W-- "+Thread.currentThread().getName()+" got notify from 'this'");
                                     // ActionListener will give a wakeup 
            }
            catch (InterruptedException e)
            {
                //e.printStackTrace();
                FileReporterHandler.verbose("W!! "+Thread.currentThread().getName()+" interrupted in wait loop; cascading throw");
                throw e;
            }

        }
        
        
        DisposeResolutionButton();
        
        if (resolution == btnDelete)
            throw new java.util.concurrent.CancellationException("Delete WID request");
        
        FileReporterHandler.verbose("C-- "+Thread.currentThread().getName()+" exits UIDRequiredSuspender.Suspend");
        
    }
    
    private synchronized void Resolve()
    {
        FileReporterHandler.verbose("C++ "+Thread.currentThread().getName()+" enters UIDRequiredSuspender.Resolve");
        
        resolution = null;
        // Pop up a UIDFrame asking for some info about this addon
        frame = new UIDFrame();
        btnOK = frame.SetCallbackOK(this);
        btnCancel = frame.SetCallbackCancel(this);
        btnDelete = frame.SetCallbackDelete(this);
        frame.setLocationRelativeTo(null);
        frame.validate();
        
        frame.identify(addon.GetName(), addon.GetAuthor());
        
        frame.setVisible(true);
       
        
        /*
        while(resolution == null)
        {
            Main.f.repaint();
            frame.repaint();
        }
        */
        FileReporterHandler.verbose("C-- "+Thread.currentThread().getName()+" exits UIDRequiredSuspender.Resolve");
        //return (resolution == btnOK);
    }

    /** Handles a performed action reported by an <tt>ActionListener</tt>
     * @param e the event that occurred
     */
    public synchronized void actionPerformed(ActionEvent e) {
        FileReporterHandler.verbose("C++ "+Thread.currentThread().getName()+" enters UIDRequiredSuspender.actionPerformed");
        if (e.getSource()==resolveBtn)
        {
            //isResolved = Resolve();
            DisableResolve();        
            Resolve();
        }
        else if (e.getSource()==btnOK)
        {
            // User clicked OK. Let's store the new UID and we're good to go
            addon.SetUID(frame.GetUID());
            resolution = btnOK;
            frame.dispose();
            EnableResolve();
            FileReporterHandler.verbose("N** "+Thread.currentThread().getName()+" broadcasts notify to 'this' (ALL)");
            this.notifyAll();
            
        }
        else if (e.getSource()==btnCancel)
        {
            resolution = btnCancel;
            frame.dispose();
            EnableResolve();
            FileReporterHandler.verbose("N** "+Thread.currentThread().getName()+" broadcasts notify to 'this' (ALL)");
            this.notifyAll();
        }
        else if (e.getSource()==btnDelete)
        {
            resolution = btnDelete;
            frame.dispose();
            EnableResolve();
            FileReporterHandler.verbose("N** "+Thread.currentThread().getName()+" broadcasts notify to 'this' (ALL)");
            this.notifyAll();
        }
        FileReporterHandler.verbose("C-- "+Thread.currentThread().getName()+" exits UIDRequiredSuspender.actionPerformed");
    }
    
    
    
    

    
}
